import abc
import datetime
import inspect
import re
import textwrap
import warnings
from pathlib import Path

import black
import enlighten
import numpy as np
import sklearn
import sklearn.cluster
import sklearn.cross_decomposition
import sklearn.feature_extraction
import sklearn.impute
import sklearn.naive_bayes
from numpy import inf, nan
from sklearn.datasets import make_classification

from autogoal import kb
from autogoal.contrib.sklearn._utils import get_input_output, is_algorithm
from autogoal.grammar import Boolean, Categorical, Continuous, Discrete
from autogoal.utils import nice_repr

from joblib import parallel_backend


try:
    import dask
    from dask.distributed import Client

    DASK_CLIENT = Client(processes=False)
    PARALLEL_BACKEND = 'dask'
except ImportError:
    PARALLEL_BACKEND = 'loky'


@nice_repr
class SklearnWrapper(metaclass=abc.ABCMeta):
    def __init__(self):
        self._mode = "train"

    def train(self):
        self._mode = "train"

    def eval(self):
        self._mode = "eval"

    def run(self, input):
        if self._mode == "train":
            return self._train(input)
        elif self._mode == "eval":
            return self._eval(input)

        raise ValueError("Invalid mode: %s" % self._mode)

    @abc.abstractmethod
    def _train(self, input):
        pass

    @abc.abstractmethod
    def _eval(self, input):
        pass


class SklearnEstimator(SklearnWrapper):
    def _train(self, input):
        X, y = input

        with parallel_backend(PARALLEL_BACKEND):
            self.fit(X, y)

        return y

    def _eval(self, input):
        X, _ = input

        with parallel_backend(PARALLEL_BACKEND):
            return self.predict(X)

    @abc.abstractmethod
    def fit(self, X, y):
        pass

    @abc.abstractmethod
    def predict(self, X):
        pass


class SklearnTransformer(SklearnWrapper):
    def _train(self, input):
        X = input

        with parallel_backend(PARALLEL_BACKEND):
            return self.fit_transform(X)

    def _eval(self, input):
        X = input

        with parallel_backend(PARALLEL_BACKEND):
            return self.transform(X)

    @abc.abstractmethod
    def fit_transform(self, X, y=None):
        pass

    @abc.abstractmethod
    def transform(self, X, y=None):
        pass


GENERATION_RULES = dict(
    LatentDirichletAllocation=dict(ignore_params=set(["evaluate_every"])),
    RadiusNeighborsClassifier=dict(ignore=True,),
    KNeighborsTransformer=dict(ignore_params=set(["metric"])),
    RadiusNeighborsTransformer=dict(ignore_params=set(["metric"])),
    LocalOutlierFactor=dict(ignore_params=set(["metric"])),
    RadiusNeighborsRegressor=dict(ignore_params=set(["metric"])),
    LabelBinarizer=dict(
        ignore_params=set(["neg_label", "pos_label"]),
        input_annotation=kb.List(kb.Category()),
    ),
    HashingVectorizer=dict(
        ignore_params=set(["token_pattern", "analyzer", "input", "decode_error"])
    ),
    SpectralBiclustering=dict(ignore_params=set(["n_components", "n_init"])),
    SpectralCoclustering=dict(ignore_params=set(["n_init"])),
    KMeans=dict(ignore_params=set(["n_init"])),
    MiniBatchKMeans=dict(ignore_params=set(["batch_size", "n_init"])),
    DictionaryLearning=dict(ignore=True),
    MiniBatchDictionaryLearning=dict(ignore=True),
    LassoLars=dict(ignore_params=["alpha"]),
    TheilSenRegressor=dict(ignore_params=["max_subpopulation"]),
    TSNE=dict(ignore=True, ignore_params=["perplexity"]),
)


def build_sklearn_wrappers():
    imports = _walk(sklearn)

    manager = enlighten.get_manager()
    counter = manager.counter(total=len(imports), unit="classes")

    path = Path(__file__).parent / "_generated.py"

    with open(path, "w") as fp:
        fp.write(
            textwrap.dedent(
                f"""
            # AUTOGENERATED ON {datetime.datetime.now()}
            ## DO NOT MODIFY THIS FILE MANUALLY

            from numpy import inf, nan

            from autogoal.grammar import Continuous, Discrete, Categorical, Boolean
            from autogoal.contrib.sklearn._builder import SklearnEstimator, SklearnTransformer
            from autogoal.kb import *
            """
            )
        )

        for cls in imports:
            counter.update()
            _write_class(cls, fp)

    black.reformat_one(path, True, black.WriteBack.YES, black.FileMode(), black.Report())

    counter.close()
    manager.stop()


def _write_class(cls, fp):
    rules = GENERATION_RULES.get(cls.__name__, {})

    if rules.get("ignore"):
        return

    ignore_args = rules.get("ignore_params", [])

    print("Generating class: %r" % cls)

    args = _get_args(cls)

    for a in ignore_args:
        args.pop(a, None)

    args = _get_args_values(cls, args)
    inputs, outputs = get_input_output(cls)
    inputs = rules.get("input_annotation", inputs)
    outputs = rules.get("output_annotation", outputs)

    if not inputs:
        warnings.warn("Cannot find correct types for %r" % cls)
        return

    s = " " * 4
    args_str = f",\n{s * 4}".join(f"{key}: {value}" for key, value in args.items())
    # self_str = f"\n{s * 4}".join(f"self.{key}={key}" for key in args)
    init_str = f",\n{s * 5}".join(f"{key}={key}" for key in args)
    input_str, output_str = repr(inputs), repr(outputs)

    base_class = (
        "SklearnEstimator" if is_algorithm(cls) == "estimator" else "SklearnTransformer"
    )

    print(cls)

    fp.write(
        textwrap.dedent(
            f"""
        from {cls.__module__} import {cls.__name__} as _{cls.__name__}

        class {cls.__name__}(_{cls.__name__}, {base_class}):
            def __init__(
                self,
                {args_str}
            ):
                {base_class}.__init__(self)
                _{cls.__name__}.__init__(
                    self,
                    {init_str}
                )

            def run(self, input: {input_str}) -> {output_str}:
               return {base_class}.run(self, input)
        """
        )
    )

    fp.flush()


def _is_algorithm(cls, verbose=False):
    if hasattr(cls, "fit"):
        return True
    else:
        if verbose:
            warnings.warn("%r doesn't have `fit`" % cls)

    if hasattr(cls, "transform"):
        return True
    else:
        if verbose:
            warnings.warn("%r doesn't have `transform`" % cls)

    return False


def _walk(module):
    imports = set()

    def _walk_p(module, visited):
        if module in visited:
            return

        visited.add(module)
        all_classes = inspect.getmembers(module, inspect.isclass)

        for name, obj in all_classes:
            if obj in imports:
                continue

            try:
                if not "sklearn" in inspect.getfile(obj):
                    continue

                print(obj)

                if isinstance(obj, type):
                    if name.endswith("CV"):
                        continue

                    if not _is_algorithm(obj):
                        continue

                    imports.add(obj)

            except Exception as e:
                print(repr(e))

        for name, inner_module in inspect.getmembers(module, inspect.ismodule):
            if not "sklearn" in inspect.getfile(module):
                continue

            print(inner_module)
            _walk_p(inner_module, visited)

    _walk_p(module, set())

    imports = sorted(imports, key=lambda c: (c.__module__, c.__name__))

    print("Finally found classes:")

    for cls in imports:
        print(cls)

    return imports


def _find_parameter_values(parameter, cls):
    documentation = []
    lines = cls.__doc__.split("\n")

    while lines:
        l = lines.pop(0)
        if l.strip().startswith(parameter):
            documentation.append(l)
            tabs = l.index(parameter)
            break

    while lines:
        l = lines.pop(0)

        if not l.strip():
            continue

        if l.startswith(" " * (tabs + 1)):
            documentation.append(l)
        else:
            break

    options = set(re.findall(r"'(\w+)'", " ".join(documentation)))
    valid = []
    invalid = []
    skip = set(["deprecated", "auto_deprecated", "precomputed"])

    for opt in options:
        opt = opt.lower()

        if opt in skip:
            continue

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if _try(cls, parameter, opt):
                valid.append(opt)
            else:
                invalid.append(opt)

    if valid:
        return Categorical(*sorted(valid))

    return None



X, y = make_classification()


def _try(cls, arg, value):
    try:
        print(f"Trying {cls}({arg}={value})... ", end="")
        cls(**{arg: value}).fit(X, y)
        print("OK")
        return True
    except:
        print("ERROR")
        return False


def _get_args(cls):
    specs = inspect.getfullargspec(cls.__init__)

    args = specs.args
    specs = specs.defaults

    if not args or not specs:
        return {}

    args = args[-len(specs) :]

    args_map = {k: v for k, v in zip(args, specs)}

    drop_args = [
        "verbose",
        "random_state",
        "n_jobs",
        "max_iter",
        "class_weight",
        "warm_start",
        "copy_X",
        "copy_x",
        "copy",
        "eps",
        "n_init",
    ]

    for arg in drop_args:
        args_map.pop(arg, None)

    print("Found args: %r" % args_map)

    return args_map


def _get_args_values(cls, args_map):
    result = {}

    for arg, value in args_map.items():
        values = _get_arg_values(arg, value, cls)
        if not values:
            continue
        result[arg] = values

    return result


def _get_arg_values(arg, value, cls):
    print(f"Computing valid values for: {arg}={value}")

    try:
        if isinstance(value, bool):
            annotation = Boolean()
        elif isinstance(value, int):
            annotation = _get_integer_values(arg, value, cls)
        elif isinstance(value, float):
            annotation = _get_float_values(arg, value, cls)
        elif isinstance(value, str):
            annotation = _find_parameter_values(arg, cls)
        else:
            annotation = None
    except:
        annotation = None

    print(f"Found annotation {arg}:{annotation}")

    return annotation


def _get_integer_values(arg, value, cls):
    if value > 0:
        min_value = 0
        max_value = 2 * value
    elif value == 0:
        min_value = -100
        max_value = 100
    else:
        return None

    # binary search for minimum value
    left = min_value
    right = value

    while left < right:
        current_value = int((left + right) / 2)
        if current_value in [left, right]:
            break

        if _try(cls, arg, current_value):
            right = current_value
        else:
            left = current_value

    min_value = right

    # binary search for maximum value
    left = value
    right = max_value

    while left < right:
        current_value = int((left + right) / 2)
        if current_value in [left, right]:
            break

        if _try(cls, arg, current_value):
            left = current_value
        else:
            right = current_value

    max_value = left

    if min_value < max_value:
        return Discrete(min=min_value, max=max_value)

    return None


def _get_float_values(arg, value, cls):
    if value in [inf, nan]:
        return None

    if value > 0:
        min_value = -10 * value
        max_value = 10 * value
    elif value == 0:
        min_value = -1
        max_value = 1
    else:
        return None

    # binary search for minimum value
    left = min_value
    right = value

    while abs(left - right) > 1e-2:
        current_value = round((left + right) / 2, 3)
        if _try(cls, arg, current_value):
            right = current_value
        else:
            left = current_value

    min_value = right

    # binary search for maximum value
    left = value
    right = max_value

    while abs(left - right) > 1e-2:
        current_value = round((left + right) / 2, 3)
        if _try(cls, arg, current_value):
            left = current_value
        else:
            right = current_value

    max_value = left

    if max_value - min_value >= 2 * value:
        return Continuous(min=min_value, max=max_value)

    return None


if __name__ == "__main__":
    build_sklearn_wrappers()



import inspect
import random
import warnings
import sys
from typing import List, Dict, Set

from ._base import Grammar, Sampler


class Symbol:
    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "Symbol(name=%r)" % (self.name)


class Production:
    def __init__(self, head: Symbol, grammar: "ContextFreeGrammar"):
        self.head = head
        self.grammar = grammar

    def to_string(
        self, code: List[str], visited: Set[Symbol], max_symbol_length: int,
    ):
        raise NotImplementedError()

    def sample(self, sampler, namespace, max_iterations):
        raise NotImplementedError()


class Empty(Production):
    def __repr__(self):
        return "Empty()"

    def to_string(
        self, code: List[str], visited: Set[Symbol], max_symbol_length: int,
    ):
        code.append(
            "%s := %s" % (("<%s>" % self.head.name).ljust(max_symbol_length), "<empty>")
        )
        visited.add(self.head)

    def sample(self, sampler, namespace, max_iterations):
        return None


class OneOf(Production):
    def __init__(self, head, grammar, *options):
        super().__init__(head, grammar)
        self._options: List[Symbol] = list(options)

    @property
    def options(self) -> List[Symbol]:
        return self._options

    def __repr__(self):
        return "OneOf(options=%r)" % self._options

    def to_string(
        self, code: List[str], visited: Set[Symbol], max_symbol_length: int,
    ):
        lhs = ("<%s>" % self.head.name).ljust(max_symbol_length)
        rhs = ["<%s>" % option.name for option in self._options]

        code.append("%s := %s" % (lhs, " | ".join(rhs)))
        visited.add(self.head)

        for symbol in self._options:
            if symbol in visited:
                continue

            self.grammar[symbol].to_string(code, visited, max_symbol_length)

    def sample(self, sampler, namespace, max_iterations):
        if max_iterations <= 0:
            raise ValueError("Max iterations exceeded")

        option = sampler.choice(self.options, handle=self.head.name)
        return self.grammar[option].sample(sampler, namespace, max_iterations-1)


class SubsetOf(Production):
    def __init__(self, head, grammar, *options, allow_empty=False):
        super().__init__(head, grammar)
        self._options: List[Symbol] = list(options)
        self._allow_empty = allow_empty

    @property
    def options(self) -> List[Symbol]:
        return self._options

    def __repr__(self):
        return "SubsetOf(options=%r)" % self._options

    def to_string(
        self, code: List[str], visited: Set[Symbol], max_symbol_length: int,
    ):
        lhs = ("<%s>" % self.head.name).ljust(max_symbol_length)
        rhs = ["<%s>" % option.name for option in self._options]

        code.append("%s := { %s }" % (lhs, " , ".join(rhs)))
        visited.add(self.head)

        for symbol in self._options:
            if symbol in visited:
                continue

            self.grammar[symbol].to_string(code, visited, max_symbol_length)

    def sample(self, sampler, namespace, max_iterations):
        if max_iterations <= 0:
            raise ValueError("Max iterations exceeded")

        while True:
            selected = []

            for option in self.options:
                if sampler.boolean(handle=self.head.name + "_" + option.name):
                    selected.append(self.grammar[option].sample(sampler, namespace, max_iterations-1))

            if len(selected) > 0 or self._allow_empty:
                break

        return selected


class Callable(Production):
    def __init__(self, head, grammar, name, **parameters):
        super().__init__(head, grammar)
        self._name = name
        self._parameters = parameters

    def __repr__(self):
        return "Callable(name=%r, parameters=%r)" % (self._name, self._parameters)

    def to_string(
        self, code: List[str], visited: Set[Symbol], max_symbol_length: int,
    ):
        lhs = ("<%s>" % self.head.name).ljust(max_symbol_length)
        rhs = [
            "%s=%s" % (key, ("<%s>" % value.name) if hasattr(value, "name") else value)
            for key, value in self._parameters.items()
        ]

        code.append("%s := %s (%s)" % (lhs, self._name, ", ".join(rhs)))
        visited.add(self.head)

        for _, symbol in self._parameters.items():
            if not isinstance(symbol, Symbol):
                continue

            if symbol in visited:
                continue

            if symbol not in self.grammar:
                continue

            self.grammar[symbol].to_string(code, visited, max_symbol_length)

    def sample(self, sampler, namespace, max_iterations):
        if max_iterations <= 0:
            raise ValueError("Max iterations exceeded")

        kwargs = {}

        for arg, symbol in self._parameters.items():
            if isinstance(symbol, Symbol):
                arg_value = self.grammar[symbol].sample(sampler, namespace, max_iterations-1)
            else:
                arg_value = symbol

            kwargs[arg] = arg_value

        obj = namespace[self._name](**kwargs)

        if hasattr(obj, "sample") and callable(obj.sample):
            obj.sample(sampler, max_iterations=max_iterations)

        return obj


class Distribution(Callable):
    def __repr__(self):
        return "Distribution(name=%r, parameters=%r)" % (self._name, self._parameters)

    def sample(self, sampler, namespace, max_iterations):
        return sampler.distribution(self._name, handle=self.head.name, **self._parameters)


class ContextFreeGrammar(Grammar):
    """Represents a CFG grammar.
    """

    def __init__(self, start: Symbol, namespace: Dict[str, type] = None):
        super(ContextFreeGrammar, self).__init__(start)
        self._namespace = {} if namespace is None else namespace
        self._productions: Dict[Symbol, Production] = {}

    @property
    def namespace(self):
        return self._namespace

    def add(self, symbol: Symbol, production: Production) -> None:
        if symbol in self:
            raise ValueError(
                "Cannot add more than once, call Grammar.replace() instead."
            )

        self._productions[symbol] = production

    def replace(self, symbol: Symbol, production: Production) -> None:
        if symbol not in self:
            raise ValueError("Symbol is not defined, call Grammar.add() instead.")

        self._productions[symbol] = production

    def __contains__(self, symbol: Symbol):
        return symbol in self._productions

    def __getitem__(self, symbol: Symbol) -> Production:
        return self._productions[symbol]

    def __repr__(self):
        return "Grammar(start=%r, productions=%r)" % (self._start, self._productions,)

    def __str__(self):
        code = []
        max_symbol_length = max(len(symbol.name) for symbol in self._productions) + 2
        self[self._start].to_string(code, set(), max_symbol_length)
        return "\n".join(code)

    def _sample(self, symbol, max_iterations, sampler):
        production = self[symbol]
        return production.sample(sampler, self._namespace, max_iterations)


def generate_cfg(cls, registry=None) -> ContextFreeGrammar:
    """
    Generates a [ContextFreeGrammar](/api/autogoal.grammar/#contextfreegrammar)
    from an annotated callable (class or function).

    ##### Parameters

    * `cls`: class or function with annotated arguments.

    ##### Returns

    * `ContextFreeGrammar`: the generated grammar.

    ##### Examples

    ```python
    >>> class MyClass:
    ...     def __init__(self, x: Discrete(1,3), y: Continuous(0,1)):
    ...         pass
    >>> grammar = generate_cfg(MyClass)
    >>> print(grammar)
    <MyClass>   := MyClass (x=<MyClass_x>, y=<MyClass_y>)
    <MyClass_x> := discrete (min=1, max=3)
    <MyClass_y> := continuous (min=0, max=1)

    ```
    """
    return _generate_cfg(cls, registry=registry)


def _generate_cfg(
    cls, grammar: ContextFreeGrammar = None, head: Symbol = None, registry = None
) -> ContextFreeGrammar:
    symbol = head or Symbol(cls.__name__)

    if grammar is None:
        grammar = ContextFreeGrammar(start=symbol)

        # pre-register all classes that are already given
        if registry:
            for clss in registry:
                grammar.namespace[clss.__name__] = clss
    elif symbol in grammar:
        return grammar

    grammar.namespace[symbol.name] = cls

    if hasattr(cls, "generate_cfg"):
        return cls.generate_cfg(grammar, symbol)

    grammar.add(symbol, Empty(symbol, grammar))
    parameters = {}

    if inspect.isclass(cls):
        signature = inspect.signature(cls.__init__)
    elif inspect.isfunction(cls):
        signature = inspect.signature(cls)
    else:
        raise ValueError("Unable to obtain signature for %r" % cls)

    for param_name, param_obj in signature.parameters.items():
        if param_name in ["self", "args", "kwargs"]:
            continue

        annotation_cls = param_obj.annotation

        if annotation_cls == inspect.Parameter.empty:
            if param_obj.default == inspect.Parameter.empty:
                raise TypeError(
                    "In <%s>: Couldn't find annotation type for %r"
                    % (cls.__name__, param_obj)
                )
            continue

        if annotation_cls == "self":
            annotation_cls = cls

        if isinstance(annotation_cls, str):
            try:
                annotation_cls = grammar.namespace[annotation_cls]
            except KeyError:
                raise ValueError("To use strings for annotations, make sure recursion hits the corresponding class first.")

        if hasattr(annotation_cls, "__name__"):
            param_symbol = Symbol(annotation_cls.__name__)
        else:
            param_symbol = Symbol("%s_%s" % (cls.__name__, param_name))

        _generate_cfg(annotation_cls, grammar, param_symbol)
        parameters[param_name] = param_symbol

    grammar.replace(symbol, Callable(symbol, grammar, cls.__name__, **parameters))
    return grammar


class Discrete:
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def __repr__(self):
        return "Discrete(min=%r, max=%r)" % (self.min, self.max)

    def generate_cfg(self, grammar, head):
        grammar.add(
            head, Distribution(head, grammar, "discrete", min=self.min, max=self.max)
        )
        return grammar


class Continuous(Discrete):
    def __repr__(self):
        return "Continuous(min=%r, max=%r)" % (self.min, self.max)

    def generate_cfg(self, grammar, head):
        grammar.add(
            head, Distribution(head, grammar, "continuous", min=self.min, max=self.max)
        )
        return grammar


class Categorical:
    def __init__(self, *options):
        self.options = list(options)

    def __repr__(self):
        options = ", ".join(repr(o) for o in self.options)
        return f"Categorical({options})"

    def generate_cfg(self, grammar, head):
        grammar.add(
            head, Distribution(head, grammar, "categorical", options=self.options)
        )
        return grammar


class Boolean:
    def __repr__(self):
        return f"Boolean()"

    @staticmethod
    def generate_cfg(grammar, head):
        grammar.add(head, Distribution(head, grammar, "boolean"))
        return grammar


class Union:
    def __init__(self, name, *clss):
        self.__name__ = name
        self.clss = list(clss)

    def __repr__(self):
        # args = ", ".join(str(s) for s in self.clss)
        return self.__name__

    def generate_cfg(self, grammar, head):
        symbol = head or Symbol(self.__name__)

        if symbol in grammar:
            return grammar

        grammar.add(symbol, Empty(symbol, grammar))
        children = []

        for child in self.clss:
            child_symbol = Symbol(child.__name__)
            children.append(child_symbol)
            _generate_cfg(child, grammar, child_symbol)

        grammar.replace(symbol, OneOf(symbol, grammar, *children))
        return grammar


class Subset:
    def __init__(self, name, *clss, allow_empty=False):
        self.__name__ = name
        self.clss = list(clss)
        self.allow_empty = allow_empty

    def __repr__(self):
        # args = ", ".join(str(s) for s in self.clss)
        return self.__name__

    def generate_cfg(self, grammar, head):
        symbol = head or Symbol(self.__name__)

        if symbol in grammar:
            return grammar

        grammar.add(symbol, Empty(symbol, grammar))
        children = []

        for child in self.clss:
            child_symbol = Symbol(child.__name__)
            children.append(child_symbol)
            _generate_cfg(child, grammar, child_symbol)

        grammar.replace(symbol, SubsetOf(symbol, grammar, *children, allow_empty=self.allow_empty))
        return grammar


class CfgInitializer:
    def __init__(self, registry=None):
        self._grammars = {}
        self._registry = registry

    def __call__(self, cls, sampler=None):
        if cls not in self._grammars:
            self._grammars[cls] =  generate_cfg(cls, self._registry)

        return self._grammars[cls].sample(sampler=sampler)



# AUTOGENERATED ON 2020-02-01 12:18:14.877438
## DO NOT MODIFY THIS FILE MANUALLY

from autogoal.grammar import Continuous, Discrete, Categorical, Boolean
from autogoal.contrib.nltk._builder import (
    NltkStemmer,
    NltkTokenizer,
    NltkLemmatizer,
    NltkTagger,
    NltkTrainedTagger,
)
from autogoal.kb._data import *
from autogoal.utils import nice_repr
from numpy import inf, nan

from nltk.stem.cistem import Cistem as _Cistem


@nice_repr
class Cistem(_Cistem, NltkStemmer):
    def __init__(self, case_insensitive: Boolean()):
        self.case_insensitive = case_insensitive
        NltkStemmer.__init__(self)
        _Cistem.__init__(self, case_insensitive=case_insensitive)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.isri import ISRIStemmer as _ISRIStemmer


@nice_repr
class ISRIStemmer(_ISRIStemmer, NltkStemmer):
    def __init__(self,):

        NltkStemmer.__init__(self)
        _ISRIStemmer.__init__(self,)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.lancaster import LancasterStemmer as _LancasterStemmer


@nice_repr
class LancasterStemmer(_LancasterStemmer, NltkStemmer):
    def __init__(self, strip_prefix_flag: Boolean()):
        self.strip_prefix_flag = strip_prefix_flag
        NltkStemmer.__init__(self)
        _LancasterStemmer.__init__(self, strip_prefix_flag=strip_prefix_flag)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.porter import PorterStemmer as _PorterStemmer


@nice_repr
class PorterStemmer(_PorterStemmer, NltkStemmer):
    def __init__(self,):

        NltkStemmer.__init__(self)
        _PorterStemmer.__init__(self,)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.rslp import RSLPStemmer as _RSLPStemmer


@nice_repr
class RSLPStemmer(_RSLPStemmer, NltkStemmer):
    def __init__(self,):

        NltkStemmer.__init__(self)
        _RSLPStemmer.__init__(self,)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.snowball import SnowballStemmer as _SnowballStemmer


@nice_repr
class SnowballStemmer(_SnowballStemmer, NltkStemmer):
    def __init__(
        self,
        language: Categorical(
            "italian",
            "portuguese",
            "hungarian",
            "english",
            "german",
            "arabic",
            "danish",
            "norwegian",
            "finnish",
            "dutch",
            "romanian",
            "russian",
            "swedish",
            "spanish",
            "french",
        ),
    ):
        self.language = language
        NltkStemmer.__init__(self)
        _SnowballStemmer.__init__(self, language=language)

    def run(self, input: Word()) -> Stem():
        return NltkStemmer.run(self, input)


from nltk.stem.wordnet import WordNetLemmatizer as _WordNetLemmatizer


@nice_repr
class WordNetLemmatizer(_WordNetLemmatizer, NltkLemmatizer):
    def __init__(self,):

        NltkLemmatizer.__init__(self)
        _WordNetLemmatizer.__init__(self,)

    def run(self, input: Word()) -> Stem():
        return NltkLemmatizer.run(self, input)


from nltk.tag.perceptron import PerceptronTagger as _PerceptronTagger


@nice_repr
class PerceptronTagger(_PerceptronTagger, NltkTrainedTagger):
    def __init__(self,):

        NltkTrainedTagger.__init__(self)
        _PerceptronTagger.__init__(self,)

    def run(self, input: List(Word())) -> List(Postag()):
        return NltkTrainedTagger.run(self, input)


from nltk.tag.sequential import AffixTagger as _AffixTagger


@nice_repr
class AffixTagger(NltkTagger):
    def __init__(
        self,
        affix_length: Discrete(min=-2, max=-6),
        min_stem_length: Discrete(min=1, max=4),
        cutoff: Discrete(min=0, max=10),
        backoff: algorithm(
            Tuple(List(List(Word())), List(List(Postag()))), List(List(Postag()))
        ),
    ):
        self.affix_length = affix_length
        self.min_stem_length = min_stem_length
        self.cutoff = cutoff
        self.backoff = backoff
        self.tagger = _AffixTagger

        self.values = dict(
            affix_length=affix_length,
            min_stem_length=min_stem_length,
            cutoff=cutoff,
            backoff=backoff,
        )

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import BigramTagger as _BigramTagger


@nice_repr
class BigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _BigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import ClassifierBasedPOSTagger as _ClassifierBasedPOSTagger


@nice_repr
class ClassifierBasedPOSTagger(NltkTagger):
    def __init__(self,):

        self.tagger = _ClassifierBasedPOSTagger
        self.values = dict()

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import TrigramTagger as _TrigramTagger


@nice_repr
class TrigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _TrigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import UnigramTagger as _UnigramTagger


@nice_repr
class UnigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _UnigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.tnt import TnT as _TnT


@nice_repr
class TnT(_TnT, NltkTrainedTagger):
    def __init__(
        self, Trained: Boolean(), N: Discrete(min=500, max=2000), C: Boolean()
    ):
        self.Trained = Trained
        self.N = N
        self.C = C
        NltkTrainedTagger.__init__(self)
        _TnT.__init__(self, Trained=Trained, N=N, C=C)

    def run(self, input: List(Word())) -> List(Postag()):
        return NltkTrainedTagger.run(self, input)


from nltk.tokenize.casual import TweetTokenizer as _TweetTokenizer


@nice_repr
class TweetTokenizer(_TweetTokenizer, NltkTokenizer):
    def __init__(
        self, preserve_case: Boolean(), reduce_len: Boolean(), strip_handles: Boolean()
    ):
        self.preserve_case = preserve_case
        self.reduce_len = reduce_len
        self.strip_handles = strip_handles
        NltkTokenizer.__init__(self)
        _TweetTokenizer.__init__(
            self,
            preserve_case=preserve_case,
            reduce_len=reduce_len,
            strip_handles=strip_handles,
        )

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.mwe import MWETokenizer as _MWETokenizer


@nice_repr
class MWETokenizer(_MWETokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _MWETokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.punkt import PunktSentenceTokenizer as _PunktSentenceTokenizer


@nice_repr
class PunktSentenceTokenizer(_PunktSentenceTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _PunktSentenceTokenizer.__init__(self,)

    def run(self, input: Document()) -> List(Sentence()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.regexp import BlanklineTokenizer as _BlanklineTokenizer


@nice_repr
class BlanklineTokenizer(_BlanklineTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _BlanklineTokenizer.__init__(self,)

    def run(self, input: Document()) -> List(Sentence()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.regexp import WhitespaceTokenizer as _WhitespaceTokenizer


@nice_repr
class WhitespaceTokenizer(_WhitespaceTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _WhitespaceTokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.regexp import WordPunctTokenizer as _WordPunctTokenizer


@nice_repr
class WordPunctTokenizer(_WordPunctTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _WordPunctTokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.sexpr import SExprTokenizer as _SExprTokenizer


@nice_repr
class SExprTokenizer(_SExprTokenizer, NltkTokenizer):
    def __init__(self, strict: Boolean()):
        self.strict = strict
        NltkTokenizer.__init__(self)
        _SExprTokenizer.__init__(self, strict=strict)

    def run(self, input: Document()) -> List(Sentence()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.simple import LineTokenizer as _LineTokenizer


@nice_repr
class LineTokenizer(_LineTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _LineTokenizer.__init__(self,)

    def run(self, input: Document()) -> List(Sentence()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.simple import SpaceTokenizer as _SpaceTokenizer


@nice_repr
class SpaceTokenizer(_SpaceTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _SpaceTokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.simple import TabTokenizer as _TabTokenizer


@nice_repr
class TabTokenizer(_TabTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _TabTokenizer.__init__(self,)

    def run(self, input: Document()) -> List(Sentence()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.toktok import ToktokTokenizer as _ToktokTokenizer


@nice_repr
class ToktokTokenizer(_ToktokTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _ToktokTokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tokenize.treebank import TreebankWordTokenizer as _TreebankWordTokenizer


@nice_repr
class TreebankWordTokenizer(_TreebankWordTokenizer, NltkTokenizer):
    def __init__(self,):

        NltkTokenizer.__init__(self)
        _TreebankWordTokenizer.__init__(self,)

    def run(self, input: Sentence()) -> List(Word()):
        return NltkTokenizer.run(self, input)


from nltk.tag.perceptron import PerceptronTagger as _PerceptronTagger


@nice_repr
class PerceptronTagger(_PerceptronTagger, NltkTrainedTagger):
    def __init__(self,):

        NltkTrainedTagger.__init__(self)
        _PerceptronTagger.__init__(self,)

    def run(self, input: List(Word())) -> List(Postag()):
        return NltkTrainedTagger.run(self, input)


from nltk.tag.sequential import BigramTagger as _BigramTagger


@nice_repr
class BigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _BigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import ClassifierBasedPOSTagger as _ClassifierBasedPOSTagger


@nice_repr
class ClassifierBasedPOSTagger(NltkTagger):
    def __init__(self,):

        self.tagger = _ClassifierBasedPOSTagger
        self.values = dict()

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import TrigramTagger as _TrigramTagger


@nice_repr
class TrigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _TrigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.sequential import UnigramTagger as _UnigramTagger


@nice_repr
class UnigramTagger(NltkTagger):
    def __init__(
        self, cutoff: Discrete(min=0, max=10),
    ):
        self.cutoff = cutoff
        self.tagger = _UnigramTagger
        self.values = dict(cutoff=cutoff)

        NltkTagger.__init__(self)

    def run(
        self, input: Tuple(List(List(Word())), List(List(Postag())))
    ) -> List(List(Postag())):
        return NltkTagger.run(self, input)


from nltk.tag.tnt import TnT as _TnT


@nice_repr
class TnT(_TnT, NltkTrainedTagger):
    def __init__(
        self, Trained: Boolean(), N: Discrete(min=500, max=2000), C: Boolean()
    ):
        self.Trained = Trained
        self.N = N
        self.C = C
        NltkTrainedTagger.__init__(self)
        _TnT.__init__(self, Trained=Trained, N=N, C=C)

    def run(self, input: List(Word())) -> List(Postag()):
        return NltkTrainedTagger.run(self, input)


__all__ = [
    "Cistem",
    "ISRIStemmer",
    "LancasterStemmer",
    "PorterStemmer",
    "RSLPStemmer",
    "SnowballStemmer",
    "WordNetLemmatizer",
    "PerceptronTagger",
    "AffixTagger",
    "BigramTagger",
    "ClassifierBasedPOSTagger",
    "TrigramTagger",
    "UnigramTagger",
    "TweetTokenizer",
    "MWETokenizer",
    "PunktSentenceTokenizer",
    "BlanklineTokenizer",
    "WhitespaceTokenizer",
    "WordPunctTokenizer",
    "SExprTokenizer",
    "LineTokenizer",
    "SpaceTokenizer",
    "TabTokenizer",
    "ToktokTokenizer",
    "TreebankWordTokenizer",
    "PerceptronTagger",
    "BigramTagger",
    "ClassifierBasedPOSTagger",
    "TrigramTagger",
    "UnigramTagger",
    "TnT",
]



from autogoal.search import SearchAlgorithm, ModelSampler
from typing import Callable


class SurrogateFunction:
    def __init__(self, func, learning_algorithm):
        self.func = func
        self.learning_algorithm = learning_algorithm


class SurrogateSearch(SearchAlgorithm):
    def __init__(
        self,
        base_search: Callable[[], SearchAlgorithm],
        estimator,
        generation_size: int = 10,
        initial_pop_size: int = 10,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.base_search = base_search
        self.generation_size = generation_size
        self.initial_pop_size = initial_pop_size

        self.training_X = []
        self.training_y = []

    def _start_generation(self):
        pass

    def _finish_generation(self, fns):
        pass

    def _build_sampler(self):
        return ModelSampler()

    def _generate(self):
        if len(self.training_X) < self.initial_pop_size:
            return super()._generate()



import numpy as np
import os
from autogoal.datasets import datapath, download


def load(max_examples=None):
    """
    Loads train and test datasets from [MEDDOCAN iberleaf 2018](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN).

    ##### Examples

    ```python
    >>> X_train, y_train, X_valid, y_valid = load()
    >>> len(X_train), len(X_valid)
    (25622, 8432)
    >>> len(y_train), len(y_valid)
    (25622, 8432)

    ```
    """

    try:
        download("meddocan_2018")
    except:
        print(
            "Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry"
        )
        raise

    path = (
        str(datapath(os.path.dirname(os.path.abspath(__file__))))
        + "/data/meddocan_2018"
    )
    train_path = path + "/train/brat"
    dev_path = path + "/dev/brat"
    test_path = path + "/test/brat"

    X_train = []
    X_test = []
    y_train = []
    y_test = []

    total = 0
    success = 0
    failed = 0

    for file in os.scandir(train_path):
        if file.name.split(".")[1] == "ann":
            text, phi = parse_text_and_tags(file.path)
            brat_corpora, text, ibo_corpora = get_tagged_tokens(text, phi)
            if compare_tags(brat_corpora, phi):
                X_train.extend(text)
                y_train.extend(ibo_corpora)

    for file in os.scandir(dev_path):
        if file.name.split(".")[1] == "ann":
            text, phi = parse_text_and_tags(file.path)
            brat_corpora, text, ibo_corpora = get_tagged_tokens(text, phi)
            if compare_tags(brat_corpora, phi):
                X_train.extend(text)
                y_train.extend(ibo_corpora)

    for file in os.scandir(test_path):
        if file.name.split(".")[1] == "ann":
            text, phi = parse_text_and_tags(file.path)
            brat_corpora, text, ibo_corpora = get_tagged_tokens(text, phi)
            if compare_tags(brat_corpora, phi):
                X_test.extend(text)
                y_test.extend(ibo_corpora)

    if max_examples is not None:
        X_train = X_train[:max_examples]
        X_test = X_test[:max_examples]
        y_train = y_train[:max_examples]
        y_test = y_test[:max_examples]

    return X_train, y_train, X_test, y_test


def parse_text_and_tags(file_name=None):
    """
    Given a file representing an annotated text in Brat format
    returns the `text` and `tags` annotated.
    """
    text = ""
    phi = []

    if file_name is not None:
        text = open(os.path.splitext(file_name)[0] + ".txt", "r").read()

        for row in open(file_name, "r"):
            line = row.strip()
            if line.startswith("T"):  # Lines is a Brat TAG
                try:
                    label = line.split("\t")[1].split()
                    tag = label[0]
                    start = int(label[1])
                    end = int(label[2])
                    value = text[start:end]
                    phi.append((tag, start, end, value))
                except IndexError:
                    print(
                        "ERROR! Index error while splitting sentence '"
                        + line
                        + "' in document '"
                        + file_name
                        + "'!"
                    )
            else:  # Line is a Brat comment
                print("\tSkipping line (comment):\t" + line)
    return (text, phi)


def get_tagged_tokens(text, tags):
    """
    convert a given text and annotations in brat format to IOB tag format

    Parameters:
    - text: raw text
    - tags: tags annotated on `text` with brat format

    output:
    tuple of identified tags in brat format from text and list of tokens tagged in IOB format
    """
    tags.sort(key=lambda x: x[1])
    offset = 0
    tagged_tokens = []

    current_tag = ""
    current_tag_end = 0
    current_tag_init = 0
    processing_token = False

    token = ""
    tag = ""

    itag = 0
    next_tag_init = tags[itag][1]

    sentences = [[]]

    for char in text:
        if processing_token and current_tag_end == offset:
            tagged_tokens.append((current_tag, current_tag_init, offset, token))

            tokens = token.split()
            if len(tokens) > 1:
                sentences[-1].append((tokens[0], tag))
                for tok in tokens[1:]:
                    sentences[-1].append((tok, "I-" + current_tag))
            else:
                sentences[-1].append((token, tag))

            token = ""
            current_tag = ""
            processing_token = False

        if not processing_token and char in [
            "\n",
            " ",
            ",",
            ".",
            ";",
            ":",
            "!",
            "?",
            "(",
            ")",
        ]:
            if token:
                sentences[-1].append((token, tag))

            if char in ["\n", ".", "!", " ?"] and len(sentences[-1]) > 1:
                sentences.append([])

            token = ""
            offset += 1
            continue

        if offset == next_tag_init:
            if token:
                if char in ["\n", " ", ",", ".", ";", ":", "!", "?", "(", ")"]:
                    sentences[-1].append((token, tag))
                else:
                    token += char
                    sentences[-1].append((token, tag))
                token = ""

            current_tag = tags[itag][0]
            current_tag_init = tags[itag][1]
            current_tag_end = tags[itag][2]
            processing_token = True

            itag += 1
            next_tag_init = tags[itag][1] if itag < len(tags) else -1

        if processing_token and current_tag:
            if not token:
                tag = "B-" + current_tag
        else:
            tag = "O"
        token += char
        offset += 1

    raw_sentences = [
        [word for word, _ in sentence] for sentence in sentences if sentence
    ]
    raw_tags = [[tag for _, tag in sentence] for sentence in sentences if sentence]
    return tagged_tokens, raw_sentences, raw_tags


def compare_tags(tag_list, other_tag_list):
    """
    compare two tags lists with the same tag format:

    (`tag_name`, `start_offset`, `end_offset`, `value`)
    """
    tags_amount = len(tag_list)

    if tags_amount != len(other_tag_list):
        print(
            "missmatch of amount of tags %d vs %d" % (tags_amount, len(other_tag_list))
        )
        return False

    tag_list.sort(key=lambda x: x[1])
    other_tag_list.sort(key=lambda x: x[1])
    for i in range(tags_amount):
        if len(tag_list[i]) != len(other_tag_list[i]):
            print("missmatch of tags format")
            return False

        for j in range(len(tag_list[i])):
            if tag_list[i][j] != other_tag_list[i][j]:
                print(
                    "missmatch of tags %s vs %s"
                    % (tag_list[i][j], other_tag_list[i][j])
                )
                return False

    return True


def get_qvals(y, predicted):
    tp = 0
    fp = 0
    fn = 0
    total_sentences = 0
    for i in range(len(y)):
        for j in range(len(y[i])):
            tag = y[i][j]
            predicted_tag = predicted[i][j]

            if tag != "O":
                if tag == predicted_tag:
                    tp += 1
                else:
                    fn += 1
            elif tag != predicted_tag:
                fp += 1
        total_sentences += 1

    return tp, fp, fn, total_sentences


def leak(y, predicted):
    """
    leak evaluation function from [MEDDOCAN iberleaf 2018](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN)
    """
    tp, fp, fn, total_sentences = get_qvals(y, predicted)
    try:
        return float(fn / total_sentences)
    except ZeroDivisionError:
        return 0.0


def precision(y, predicted):
    """
    precision evaluation function from [MEDDOCAN iberleaf 2018](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN)
    """
    tp, fp, fn, total_sentences = get_qvals(y, predicted)
    try:
        return tp / float(tp + fp)
    except ZeroDivisionError:
        return 0.0


def recall(y, predicted):
    """
    recall evaluation function from [MEDDOCAN iberleaf 2018](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN)
    """
    tp, fp, fn, total_sentences = get_qvals(y, predicted)
    try:
        return tp / float(tp + fn)
    except ZeroDivisionError:
        return 0.0


def F1_beta(y, predicted, beta=1):
    """
    F1 evaluation function from [MEDDOCAN iberleaf 2018](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN)
    """
    p = precision(predicted, y)
    r = recall(predicted, y)
    try:
        return (1 + beta ** 2) * ((p * r) / (p + r))
    except ZeroDivisionError:
        return 0.0
        pass


def basic_fn(y, predicted):
    correct = 0
    total = 0
    for i in range(len(y)):
        for j in range(len(y[i])):
            total += 1

            _, tag = y[i][j]
            _, predicted_tag = predicted[i][j]
            correct += 1 if tag == predicted_tag else 0

    return correct / total


def test_meddocan():
    X_train, X_valid, y_train, y_valid = load()

    assert len(X_train) == len(y_train)
    assert len(X_valid) == len(y_valid)

    assert all(isinstance(x, list) for x in X_train)
    assert all(isinstance(x, list) for x in X_valid)

    assert all(len(x) > 0 for x in X_train)
    assert all(len(x) > 0 for x in X_valid)

    for xi, yi in zip(X_train, y_train):
        assert len(xi) == len(yi)

    for xi, yi in zip(X_valid, y_valid):
        assert len(xi) == len(yi)



"""
This module generates a random dataset useful for quickly testing the interface of AutoGOAL methods.
"""

import numpy as np


def load(samples=100, classes=2, features=10, seed=None):
    """
    Create a random X,y pair.

    ##### Examples

    ```python
    >>> X, y = load(samples=4, features=2, seed=0)
    >>> print(X)
    [[0.5488135  0.71518937]
     [0.60276338 0.54488318]
     [0.4236548  0.64589411]
     [0.43758721 0.891773  ]]
    >>> y
    array([0, 0, 0, 1])

    ```
    """

    if seed is not None:
        np.random.seed(seed)

    X = np.random.random((samples, features))
    y = np.random.randint(0, classes, samples).astype(str)

    return X, y


import os
import pandas as pd
import numpy as np

from autogoal.datasets import datapath, download


def load_raw(max_examples=None):
    """
    Loads the train and test datasets for the [HAHA 2019 corpus](https://www.fing.edu.uy/inco/grupos/pln/haha/index.html#data)
    as Pandas dataframes.

    ##### Examples

    ```python
    >>> train, test = load_raw()
    >>> len(train), len(test)
    (24000, 6000)
    >>> train.columns
    Index(['id', 'text', 'is_humor', 'votes_no', 'votes_1', 'votes_2', 'votes_3',
           'votes_4', 'votes_5', 'funniness_average'],
          dtype='object')
    >>> train["funniness_average"].mean()
    2.0464498676235694

    ```
    """

    download("haha_2019")

    train_df = pd.read_csv(datapath("haha_2019") / "haha_2019_train.csv")
    test_df = pd.read_csv(datapath("haha_2019") / "haha_2019_test_gold.csv")

    if max_examples is not None:
        train_df = train_df[:max_examples]
        test_df = test_df[:max_examples]

    return train_df, test_df


def load(target='is_humor', max_examples=None):
    """
    Loads the train and test datasets for the [HAHA 2019 corpus](https://www.fing.edu.uy/inco/grupos/pln/haha/index.html#data)
    as lists of texts and target values.

    ##### Arguments

    * `target`: Which column to use for target. Default is `"is_humor"` which can be used for binary classification.
                Another option is `"funniness_average"` which can be used for regression.

    ##### Examples

    Loading with classification targets:

    ```python
    >>> X_train, y_train, X_test, y_test = load()
    >>> print(X_train[13])
    Leí que la falta de sexo trae consigo una notable mejora en el léxico. Me quedo absorto ante tal afirmación carente de raciocinio.
    >>> y_train[13]
    1

    ```

    Loading with regression targets:

    ```python
    >>> X_train, y_train, X_test, y_test = load(target="funniness_average")
    >>> print(X_train[13])
    Leí que la falta de sexo trae consigo una notable mejora en el léxico. Me quedo absorto ante tal afirmación carente de raciocinio.
    >>> y_train[13]
    3.25

    ```

    Loading a subset of the dataset:

    ```python
    >>> Xtrain, Xtest, ytrain, ytest = load(max_examples=100)
    >>> len(Xtrain), len(Xtest), len(ytrain), len(ytest)
    (100, 100, 100, 100)

    ```

    """

    train_df, test_df = load_raw(max_examples)
    X_train = list(train_df['text'])
    y_train = list(train_df[target])
    X_test = list(test_df['text'])
    y_test = list(test_df[target])

    return X_train, np.asarray(y_train),  X_test, np.asarray(y_test)



# coding: utf-8

import os
import numpy as np


from autogoal.datasets import download, datapath
from sklearn.feature_extraction import DictVectorizer


def _parse(x):
    return int(x) if x.isdigit() else x

def load(max_examples=None):
    download("german_credit")

    f = open(datapath("german_credit") / "german.data", "r")

    X = []
    y = []

    for i in f.readlines():

        if max_examples and len(X) >= max_examples:
            break

        clean_line = i.strip().split()

        line = {'feature_%i'% i : _parse(v) for i,v in enumerate(clean_line[:-1])}

        X.append(line)
        y.append(int(clean_line[-1]) == 2)

    return DictVectorizer(sparse=False).fit_transform(X), np.asarray(y)



import random


def load(max_examples=None):
    from nltk.corpus import movie_reviews

    sentences = []
    classes = []

    ids = list(movie_reviews.fileids())
    random.shuffle(ids)

    for fd in ids:
        if fd.startswith("neg/"):
            cls = "neg"
        else:
            cls = "pos"

        fp = movie_reviews.open(fd)
        sentences.append(fp.read())
        classes.append(cls)

        if max_examples and len(classes) >= max_examples:
            break

    return sentences, classes


def make_fn(test_size=0.25, examples=None):
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import accuracy_score

    X, y = load(examples)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size)

    def fitness_fn(pipeline):
        pipeline.send("train")
        pipeline.run((X_train, y_train))
        pipeline.send("eval")
        _, y_pred = pipeline.run((X_test, [None] * len(y_test)))

        return accuracy_score(y_test, y_pred)

    return fitness_fn



import uuid
import abc
import functools
import warnings
import json
import collections
import numpy as np

from typing import List
from autogoal.search import Logger
from autogoal.utils import nice_repr
from autogoal import sampling
# from sklearn.feature_extraction import DictVectorizer


class DatasetFeatureLogger(Logger):
    def __init__(
        self,
        X,
        y=None,
        extractor=None,
        output_file="metalearning.json",
        problem_features=None,
        environment_features=None,
    ):
        self.extractor = extractor or DatasetFeatureExtractor()
        self.X = X
        self.y = y
        self.run_id = str(uuid.uuid4())
        self.output_file = output_file
        self.problem_features = problem_features or {}
        self.environment_features = environment_features or {}

    def begin(self, generations, pop_size):
        self.dataset_features_ = self.extractor.extract_features(self.X, self.y)

    def eval_solution(self, solution, fitness):
        if not hasattr(solution, "sampler_"):
            raise ("Cannot log if the underlying algorithm is not PESearch")

        sampler = solution.sampler_

        features = {k: v for k, v in sampler._updates.items() if isinstance(k, str)}
        feature_types = {k: repr(v) for k, v in sampler._model.items() if k in features}

        info = SolutionInfo(
            uuid=self.run_id,
            fitness=fitness,
            problem_features=dict(self.dataset_features_, **self.problem_features),
            environment_features=dict(self.environment_features),
            pipeline_features=features,
            feature_types=feature_types,
        ).to_dict()

        with open(self.output_file, "a") as fp:
            fp.write(json.dumps(info) + "\n")


class DatasetFeatureExtractor:
    def __init__(self, features_extractors=None):
        self.feature_extractors = list(features_extractors or _EXTRACTORS)

    def extract_features(self, X, y=None):
        features = {}

        for extractor in self.feature_extractors:
            features.update(**extractor(X, y))

        return features


_EXTRACTORS = []


def feature_extractor(func):
    @functools.wraps(func)
    def wrapper(X, y=None):
        try:
            result = func(X, y)
        except:
            result = None
            # raise

        return {func.__name__: result}

    _EXTRACTORS.append(wrapper)
    return wrapper


# Feature extractor methods


@feature_extractor
def is_supervised(X, y=None):
    return y is not None


@feature_extractor
def dimensionality(X, y=None):
    d = 1

    for di in X.shape[1:]:

        d *= di

    return d


@feature_extractor
def training_examples(X, y=None):
    try:
        return X.shape[0]
    except:
        return len(X)


@feature_extractor
def has_numeric_features(X, y=None):
    return any([xi for xi in X[0] if isinstance(xi, (float, int))])


@feature_extractor
def numeric_variance(X, y=None):
    return X.std()


@feature_extractor
def average_number_of_words(X, y=None):
    return sum(len(sentence.split(" ")) for sentence in X) / len(X)


@feature_extractor
def has_text_features(X, y=None):
    return isinstance(X[0], str)


@nice_repr
class SolutionInfo:
    def __init__(
        self,
        uuid: str,
        problem_features: dict,
        pipeline_features: dict,
        environment_features: dict,
        feature_types: dict,
        fitness: float,
    ):
        self.problem_features = problem_features
        self.pipeline_features = pipeline_features
        self.environment_features = environment_features
        self.feature_types = feature_types
        self.fitness = fitness
        self.uuid = uuid

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(d):
        return SolutionInfo(**d)


class LearnerMedia:
    def __init__(self, problem, solutions: List[SolutionInfo], beta=1):
        self.solutions = solutions
        self.problem = problem
        self.beta = beta

    def initialize(self):
        raise NotImplementedError("We need to refactor to not depend on DictVectorizer")

        self.best_fitness = collections.defaultdict(lambda: 0)
        self.all_features = {}

        for i in self.solutions:
            self.best_fitness[i.uuid] = max(self.best_fitness[i.uuid], i.fitness)

            for feature in i.pipeline_features:
                self.all_features[feature] = None

        self.vect = DictVectorizer(sparse=False)
        self.vect.fit([self.problem])

        self.weights_solution = self.calculate_weight_examples(self.solutions)

    def compute_all_features(self):
        self.initialize()

        for feature in list(self.all_features):
            self.all_features[feature] = self.compute_feature(feature)
            print(feature, "=", self.all_features[feature])

    def compute_feature(self, feature):
        """Select for training all solutions where is used the especific feature.
    
        Predict the media of the parameter value.
        """
        # find the relevant solutions, that contain the production to predict
        important_solutions = []
        feature_prototype = None

        for i, w in zip(self.solutions, self.weights_solution):
            if feature in i.pipeline_features:
                for value in i.pipeline_features[feature]:
                    important_solutions.append((value, w))

                if feature_prototype is None:
                    feature_prototype = eval(
                        i.feature_types[feature], sampling.__dict__, {}
                    )

        if feature_prototype is None:
            return None

        return feature_prototype.weighted(important_solutions)

    def calculate_weight_examples(self, solutions: List[SolutionInfo]):
        """Calcule a weight of each example considering the fitness and the similariti with the
        actual problem 
        """
        # met = fitness * (similarity)^beta
        # métrica utilizada en active learning para combinar informativeness with representativeness

        weights = []

        for info in solutions:
            # normalize fitness
            info.fitness = self.normalize_fitness(info)

            if info.fitness == 0:
                continue

            # calculate similarity
            sim = self.similarity_cosine(info.problem_features)
            # calculate metric for weight
            weights.append(info.fitness * (sim) ** self.beta)

        return weights

    def normalize_fitness(self, info: SolutionInfo):
        """Normalize the fitness with respect to the best solution in the problem where that solution is evaluated
        """
        return info.fitness / self.best_fitness[info.uuid]

    def similarity_cosine(self, other_problem):
        """Caculate the cosine similarity for a particular solution problem(other problem) 
        and the problem analizing
        """
        x = self.vect.transform(other_problem)[0]
        y = self.vect.transform(self.problem)[0]

        return np.dot(x, y) / (np.dot(x, x) ** 0.5 * np.dot(y, y) ** 0.5)

    def similarity_learning(self, other_problem):
        """ Implementar una espicie de encoding para los feature de los problemas
        """
        raise NotImplementedError()



import sys
import streamlit as st
import textwrap
import networkx as nx
import pandas as pd
import altair as alt
import nx_altair as nxa
import inspect

from autogoal.kb import build_pipeline_graph
from matplotlib import pyplot as plt


@st.cache(allow_output_mutation=True)
def eval_code(code, *variables):
    locals_dict = {}
    exec(code, globals(), locals_dict)

    if len(variables) == 0:
        return None
    if len(variables) == 1:
        return locals_dict[variables[0]]
    else:
        return [locals_dict[var] for var in variables]


class Demo:
    def __init__(self):
        self.main_sections = {
            "Intro": self.intro,
            "High-Level API": self.high_level,
            "Pipelines": self.build_pipelines,
        }

    def intro(self):
        st.write("# AutoGOAL Demos")

        st.write(
            """
            Welcome to the AutoGOAL Demo. In the left sidebar
            you will find all the available demos and additional
            controls specific to each of them.

            AutoGOAL is a framework in Python for automatically finding the best way to solve a given task.
            It has been designed mainly for automatic machine learning~(AutoML)
            but it can be used in any scenario where several possible strategies are available 
            to solve a given computational task.
            """
        )

        st.write(
            """
        ## About this demo

        The purpose of this demo application is to showcase the main use cases of AutoGOAL.
        Keep in mind that AutoGOAL is a software library, i.e., meant to be used from source code.
        This demo serves as an interactive and user-friendly introduction to the library, but it
        is in no case a full-featured AutoML application.

        There are two sections to showcase different components of AutoGOAL.
        You can switch sections in the left sidebar.
        
        * The **High-Level API** section presents the public interface of AutoGOAL
        in several datasets.
        * The **Pipelines** section shows the internal components of AutoGOAL and allows
        to explore the possible pipelines.
        """
        )

        st.write("""
        ## Running the code

        To execute this demo on your own infrastructure, you need AutoGOAL's docker image.
        There are two images available, without and without GPU support.

        Download the corresponding Docker image:

            docker pull autogoal/autogoal

        """)

        st.write(
            """
            Launch a Docker container.

                docker run --rm -p 8501:8501 autogoal/autogoal

            Navigate to <http://localhost:8501>.
            """
        )

    def high_level(self):
        st.write("# High-Level API")
        st.write(
            """
            AutoGOAL is first and foremost a framework for Automatic Machine Learning.
            With a few simple lines of code, you can quickly find a close to optimal
            solution for classic machine learning problems.
            """
        )

        from autogoal import datasets

        dataset_descriptions = {
            "cars": """
                [Cars](https://archive.ics.uci.edu/ml/datasets/Car+Evaluation)
                is a low-dimensionality supervised problem with 21 one-hot encoded features.
                """,
            "german_credit": """
                [German Credit](https://archive.ics.uci.edu/ml/datasets/Statlog+%28German+Credit+Data%29)
                is a low-dimensionality supervised problem with 20 categorical or numerical features.
                """,
            "abalone": """
                [Abalone](https://archive.ics.uci.edu/ml/datasets/Abalone)
                is a low-dimensionality supervised problem with 8 categorical or numerical features.
                """,
            "shuttle": """
                [Shuttle](https://archive.ics.uci.edu/ml/datasets/Statlog+(Shuttle))
                is a low-dimensionality supervised problem with 9 numerical features.
                """,
            "yeast": """
                [Yeast](https://archive.ics.uci.edu/ml/datasets/Yeast)
                is a low-dimensionality supervised problem with 9 numerical features.
                """,
            "dorothea": """
                [Dorothea](https://archive.ics.uci.edu/ml/datasets/dorothea)
                is a high-dimensionality sparse supervised problem with 100,000 numerical features.
                """,
            "gisette":  """
                [Gisette](https://archive.ics.uci.edu/ml/datasets/Gisette)
                is a high-dimensionality sparse supervised problem with 5,000 numerical features.
                """,
            "haha": """
                [HAHA 2019](https://www.fing.edu.uy/inco/grupos/pln/haha/index.html#data) is
                a text classification problem with binary classes in Spanish.
                """,
            "meddocan": """
                [MEDDOCAN 2019](https://github.com/PlanTL-SANIDAD/SPACCC_MEDDOCAN) is
                an entity recognition problem in Spanish medical documents.
                """,
        }

        override_types = {
            'german_credit': ("MatrixContinuousDense()", "CategoricalVector()"),
            'dorothea': ("MatrixContinuousSparse()", "CategoricalVector()"),
            'gisette': ("MatrixContinuousSparse()", "CategoricalVector()"),
            'haha': ("List(Sentence())", "CategoricalVector()"),
            'meddocan': ("List(List(Word()))", "List(List(Postag()))"),
        }

        st.write(
            """Let's start by selecting one of the example datasets.
            These are sample datasets which are automatically downloaded by AutoGOAL,
            and can be used to benchmark new algorithms and showcase AutoML tools.
            """
        )

        dataset = st.selectbox("Select a dataset", list(dataset_descriptions))

        st.write(
            dataset_descriptions[dataset] + "Here is the code to load this dataset."
        )

        code = textwrap.dedent(
            f"""
            from autogoal.datasets import {dataset}

            X, y, *_ = {dataset}.load()
        """
        )
        st.code(code)

        X, y = eval_code(code, "X", "y")

        if st.checkbox("Preview data"):
            try:
                l = len(X)
            except:
                l = X.shape[0]

            head = st.slider("Preview N first items", 0, l, 5)
            if isinstance(X, list):
                st.write(X[:head])
            else:
                st.write(X[:head, :])
            
            st.write(y[:head])

        st.write(
            """
            The next step is to instantiate an AutoML solver and run it on this problem.
            The `AutoML` class provides a black-box interface to AutoGOAL.
            You can tweak the most important parameters at the left sidebar, even though
            sensible defaults are provided for all the parameters.
            """
        )

        st.sidebar.markdown("### AutoML parameters")
        iterations = st.sidebar.number_input("Number of iterations", 1, 10000, 100)
        global_timeout = st.sidebar.number_input(
            "Global timeout (seconds)", 1, 1000, 60
        )
        pipeline_timeout = st.sidebar.number_input(
            "Timeout per pipeline (seconds)", 1, 1000, 5
        )

        from autogoal.contrib.streamlit import StreamlitLogger

        if dataset in override_types:
            input_type, output_type = override_types[dataset]
            types_code = f"""
                input={input_type},
                output={output_type},
            """
            
            st.info(f"""
            In most cases AutoGOAL can automatically infer the input and output type
            from the dataset. Sometimes, such as with `{dataset}`, the user will need to provide them
            explicitely.
            """)
        else:
            types_code = ""


        code = textwrap.dedent(f"""
            from autogoal.kb import *
            from autogoal.ml import AutoML

            automl = AutoML(
                errors="ignore",  # ignore exceptions (e.g., timeouts)
                search_iterations={iterations}, # total iterations
                search_kwargs=dict(
                    search_timeout={global_timeout}, # max time in total (approximate)
                    evaluation_timeout={pipeline_timeout}, # max time per pipeline (approximate)
                ), {types_code}
            )
            """
        )

        st.code(code)
        automl = eval_code(code, "automl")

        st.write(
            """
            Click run to call the `fit` method. Keep in mind that many of these pipelines can be 
            quite computationally heavy and both the hyperparameter configuration as well as the
            infrastructure where this demo is running might not allow for the best pipelines to execute.
            """
        )

        st.code("automl.fit(X, y)", language="Python")

        if st.button("Run it!"):
            automl.fit(X, y, logger=StreamlitLogger())

        st.write(
            """
            ## Next steps

            Take a look at the remaining examples in the sidebar.
            """
        )

    def build_pipelines(self):
        st.write("# Pipelines")

        st.write(
            "This example illustrates how AutoGOAL automatically builds "
            "a graph of pipelines for different problems settings."
        )

        from autogoal.kb._data import DATA_TYPES

        types_str = [cls.__name__ for cls in DATA_TYPES]

        st.write(
            """
            AutoGOAL pipeline discovery is based on a hierarchy of semantic datatypes.
            Each type represents a semantic datum that can be used in a machine learning algorithm,
            from matrices and vectors to sentences, entities and and images.

            The following picture shows all available semantic data types.
            You can click the top right corner to enlarge.
            """
        )

        st.image("/code/docs/guide/datatypes.png", use_column_width=True)

        from autogoal.contrib import find_classes

        all_classes = {k.__name__: k for k in find_classes()}

        st.write(
            f"""
            ## Algorithm Library

            AutoGOAL automatically builds pipelines by selecting from a wide range of algorithms
            implemented in `contrib` modules.
            The list of all available algorithms is shown here.

            There are a total of **{len(all_classes)}** algorithms implemented.
            Select one to display some information.
            """)


        class_name = st.selectbox("Select an algorithm", list(all_classes))
        class_type = all_classes[class_name]

        st.write(f"### {class_type.__module__}.{class_name}")

        run_signature = inspect.signature(class_type.run)
        st.write(f"**Input type**: {run_signature.parameters['input'].annotation}")
        st.write(f"**Output type**: {run_signature.return_annotation}")
        
        st.write("#### Parameters")
        params = []
        for name, param in inspect.signature(class_type.__init__).parameters.items():
            if name == 'self':
                continue

            params.append(f"* **{name}**: {param.annotation}")
        st.write("\n".join(params))

        st.write("## Pipeline Builder")

        st.write(
            """
            AutoGOAL can automatically build pipelines given a desired input and output
            value. It uses the annotations of the `run` method of each algorithm to detect
            which algorithms can be connected.

            In the following section, you can select a desired input and output types and 
            explore the pipelines that AutoGOAL discovers.
            In the left sidebar you can fine-tune the input value, e.g., make it a list
            of elements instead of a single element.
            """
        )


        st.sidebar.markdown("### Configure input and output types")
        list_input = st.sidebar.number_input("Input list (level)", 0, 3, 1)
        list_output = st.sidebar.number_input("Output list (level)", 0, 3, 0)
        tuples = st.sidebar.checkbox("Is supervised (use Tuple in input)", True)
        
        input_type = st.selectbox(
            "Select an input type", types_str, types_str.index('Sentence')
        )

        output_type = st.selectbox(
            "Select and output type", types_str, types_str.index('CategoricalVector')
        )

        input_type = input_type + "()"
        for i in range(list_input):
            input_type = f"List({input_type})"

        output_type = output_type + "()"
        for i in range(list_output):
            input_type = f"List({output_type})"

        if tuples:
            input_type = f"Tuple({input_type}, {output_type})"

        st.write(f"#### Defined input type:  `{input_type}`")
        st.write(f"#### Defined output type: `{output_type}`")

        st.write(
            """
            The following code uses explicitely AutoGOAL's pipeline discovery engine
            to find all the pipelines that can be constructed from the desired
            input to the desired output.
            """
        )

        code = textwrap.dedent(
            f"""
            from autogoal.kb import *
            from autogoal.kb import build_pipelines
            from autogoal.contrib import find_classes

            # explicitly build the graph of pipelines
            space = build_pipelines(
                input={input_type},
                output={output_type},
                registry=find_classes(),
            )
            """
        )

        st.code(code)

        try:
            space = eval_code(code, "space")
        except Exception as e:
            if "No pipelines can be constructed" in str(e):
                st.error(str(e))
                st.info("Try changing the input and output type or select **Is supervised** in the left sidebar.")
                return
            
            raise

        st.write(
            """
            ### The Pipelines Graph
            
            This is the graph that represents all the posible pipelines find by AutoGOAL.
            Each node in this graph is an algorithm from the _Algorithm Library_ that is
            compatible with the input and output types of its neighbors.
            Any path from the top to the bottom of the graph represents a valid pipeline.
            """)
            
        graph = nx.DiGraph()

        def get_node_repr(node):
            try:
                return get_node_repr(node.inner)
            except:
                return dict(
                    label=str(node).split(".")[-1], module=node.__module__.split("_")[0]
                )

        for node in space.graph.nodes:
            attrs = get_node_repr(node)
            graph.add_node(attrs["label"], **attrs)

        for u, v in space.graph.edges:
            graph.add_edge(get_node_repr(u)["label"], get_node_repr(v)["label"])

        pos = nx.nx_pydot.pydot_layout(graph, prog="dot", root=space.Start)
        chart = (
            nxa.draw_networkx(graph, pos=pos, node_color="module", node_tooltip="label")
            .properties(height=500)
            .interactive()
        )

        st.altair_chart(chart, use_container_width=True)

        st.write(
            """
            ### Example Pipeline
            
            Here is an example pipeline that has been randomly sampled from the previous graph.
            You can try different samples. Notice how not only the nodes (algorithms) that participate
            in the pipeline are different each time, but also their internal hyperparameters change.
            
            When sampling a pipeline from the graph AutoGOAL samples all the internal
            hyperparameters as defined by the constructor.
            When these hyperparameters have complex values (e.g., an algorithm per-se), AutoGOAL
            recursively samples instances of the internal algorithms, and so on.
            """)

        st.code(space.sample())

        st.button("Sample another pipeline")

    def run(self):
        main_section = st.sidebar.selectbox("Section", list(self.main_sections))
        self.main_sections[main_section]()


demo = Demo()
demo.run()



#

import warnings

import numpy as np
import scipy.sparse as sp


class String(str):
    def __new__(cls, x):
        return str.__new__(cls, x)


def _get_class_name(cls):
    return str(cls).split(".")[-1]


def combine_types(*types):
    if len(types) == 1:
        return types[0]

    types = set(types)

    if types == {kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()}:
        return kb.MatrixContinuous()

    return None


def is_matrix_continuous_dense(obj):
    """Determine if `obj` is a continuous dense matrix (i.e., NxM floats).

    Examples:

    >>> is_matrix_continuous_dense([[0, 1],[0, 1]])
    True
    >>> is_matrix_continuous_dense([[0,1]])
    True
    >>> is_matrix_continuous_dense([0,1])
    False
    >>> is_matrix_continuous_dense(np.random.rand(10,10))
    True
    >>> is_matrix_continuous_dense(np.random.rand(10))
    False

    """
    try:
        obj = np.asarray(obj)
        return len(obj.shape) == 2
    except:
        return False


def is_matrix_continuous_sparse(obj):
    """Determine if `obj` is a continuous sparse matrix (i.e., NxM floats).

    Examples:

    >>> is_matrix_continuous_sparse(sp.rand(10,10))
    True
    >>> is_matrix_continuous_sparse([[0,1]])
    False
    >>> is_matrix_continuous_sparse([0,1])
    False
    >>> is_matrix_continuous_sparse(np.random.rand(10,10))
    False
    >>> is_matrix_continuous_sparse(np.random.rand(10))
    False

    """
    try:
        sp.rand
        return sp.issparse(obj) and len(obj.shape) == 2
    except:
        return False


def is_categorical(obj):
    """Determines if `obj` is a sequence of categories (integer or string)

    Examples:

    >>> is_categorical(['A'] * 5 + ['B'] * 5)
    True
    >>> is_categorical(np.asarray(list('ABCABCABC')))
    True

    """
    try:
        obj = np.asarray(obj)
        assert len(obj.shape) == 1

        original_length = len(obj)
        obj = set(obj)

        a = len(obj) < max(0.1 * original_length, 10)
        b = all(isinstance(x, (str, int, np.int64, np.int32)) for x in obj)
        return a and b
    except:
        return False


def is_continuous(obj):
    """Determines if `obj` is a sequence of float values

    Examples:

    >>> is_continuous(np.random.rand(10))
    True
    >>> is_continuous(np.random.rand(10,10))
    False

    """
    try:
        obj = np.asarray(obj)
        assert len(obj.shape) == 1

        return not all(obj.astype(int) == obj)
    except:
        return False


def is_discrete(obj):
    """Determines if `obj` is a sequence of integer values

    Examples:

    >>> is_discrete(np.random.randint(0,1,(10,)))
    True
    >>> is_discrete(np.random.rand(10))
    False

    """
    try:
        obj = np.asarray(obj)
        assert len(obj.shape) == 1

        return all(obj.astype(int) == obj)
    except:
        return False


def is_string_list(obj):
    """Determines if `obj` is a sequence of strings.

    Examples:

    >>> is_string_list(['hello world', 'another sentence'])
    True
    >>> is_string_list(np.random.rand(10))
    False

    """
    try:
        obj = np.asarray(obj)
        assert len(obj.shape) == 1

        original_length = len(obj)
        obj = set(obj)

        return len(obj) > 0.1 * original_length and all(isinstance(x, str) for x in obj)
    except:
        return False


from autogoal import kb


DATA_RESOLVERS = {
    kb.MatrixContinuousDense(): is_matrix_continuous_dense,
    kb.MatrixContinuousSparse(): is_matrix_continuous_sparse,
    kb.CategoricalVector(): is_categorical,
    kb.ContinuousVector(): is_continuous,
    kb.List(kb.Sentence()): is_string_list,
}


DATA_TYPE_EXAMPLES = {
    kb.MatrixContinuousDense(): np.random.rand(10, 10),
    kb.MatrixContinuousSparse(): sp.rand(10, 10),
    kb.CategoricalVector(): np.asarray(["A"] * 5 + ["B"] * 5),
    kb.ContinuousVector(): np.random.rand(10),
    kb.DiscreteVector(): np.random.randint(0, 10, (10,), dtype=int),
    kb.List(kb.Sentence()): ["abc bcd def feg geh hij jkl lmn nop pqr"] * 10,
}


def is_algorithm(cls, verbose=False):
    if hasattr(cls, "fit") and hasattr(cls, "predict"):
        return "estimator"
    else:
        if verbose:
            warnings.warn("%r doesn't have `fit`" % cls)

    if hasattr(cls, "transform"):
        return "transformer"
    else:
        if verbose:
            warnings.warn("%r doesn't have `transform`" % cls)

    return False


def is_classifier(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an sklearn classifier.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression, LinearRegression
    >>> is_classifier(LogisticRegression)
    (True, (Tuple(MatrixContinuous(), CategoricalVector()), CategoricalVector()))
    >>> is_classifier(LinearRegression)
    (False, None)

    """
    if not is_algorithm(cls, verbose=verbose):
        return False, None

    inputs = []

    for input_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]
            y = DATA_TYPE_EXAMPLES[kb.CategoricalVector()]

            clf = cls()
            clf.fit(X, y)
            y = clf.predict(X)

            assert is_categorical(y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (kb.Tuple(inputs, kb.CategoricalVector()), kb.CategoricalVector())
    else:
        return False, None


def is_regressor(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an sklearn regressor.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression, LinearRegression
    >>> is_regressor(LogisticRegression)
    (False, None)
    >>> is_regressor(LinearRegression)
    (True, (Tuple(MatrixContinuous(), ContinuousVector()), ContinuousVector()))

    """
    if not is_algorithm(cls, verbose=verbose):
        return False, None

    inputs = []

    for input_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]
            y = DATA_TYPE_EXAMPLES[kb.ContinuousVector()]

            clf = cls()
            clf.fit(X, y)
            y = clf.predict(X)

            assert is_continuous(y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (kb.Tuple(inputs, kb.ContinuousVector()), kb.ContinuousVector())
    else:
        return False, None


def is_clusterer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an sklearn clustering algorithm.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression, LinearRegression
    >>> is_clusterer(LogisticRegression)
    (False, None)
    >>> is_clusterer(LinearRegression)
    (False, None)
    >>> from sklearn.cluster import KMeans
    >>> is_clusterer(KMeans)
    (True, (MatrixContinuous(), DiscreteVector()))

    """
    if not is_algorithm(cls, verbose=verbose):
        return False, None

    inputs = []

    for input_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            clf = cls()
            y = clf.fit_predict(X)

            assert is_discrete(y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, kb.DiscreteVector())
    else:
        return False, None


def is_transformer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an sklearn general transformer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.feature_extraction.text import CountVectorizer
    >>> is_transformer(CountVectorizer)
    (True, (List(Sentence()), MatrixContinuousSparse()))
    >>> from sklearn.decomposition.pca import PCA
    >>> is_transformer(PCA)
    (True, (MatrixContinuousDense(), MatrixContinuousDense()))

    """
    if not is_algorithm(cls, verbose=verbose):
        return False, None

    allowed_inputs = set()
    allowed_outputs = set()

    for input_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse(), kb.List(kb.Sentence())]:
        for output_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse(), kb.List(kb.Sentence())]:
            try:
                X = DATA_TYPE_EXAMPLES[input_type]

                clf = cls()
                X = clf.fit_transform(X)

                assert is_data_type(X, output_type)

                allowed_inputs.add(input_type)
                allowed_outputs.add(output_type)
            except Exception as e:
                if verbose:
                    warnings.warn(str(e))

    if len(allowed_outputs) != 1:
        return False, None

    inputs = combine_types(*allowed_inputs)

    if allowed_inputs:
        return True, (inputs, list(allowed_outputs)[0])
    else:
        return False, None


def is_data_type(X, data_type):
    return DATA_RESOLVERS[data_type](X)

IO_TYPE_HANDLER = [
    is_classifier,
    is_regressor,
    is_clusterer,
    is_transformer
]

def get_input_output(cls, verbose=False):
    for func in IO_TYPE_HANDLER:
        matches, types = func(cls, verbose=verbose)
        if matches:
            return types

    return None, None

def solve_type(obj):
    for type_, resolver in DATA_RESOLVERS.items():
        if resolver(obj):
            return type_

    raise ValueError("Unresolved type for %r" % obj)



import types
import inspect
import pprint

from typing import Mapping
from autogoal.grammar import Symbol, Union, Empty, Subset

from scipy.sparse.base import spmatrix
from numpy import ndarray


def algorithm(input_type, output_type):
    def run_method(self, input: input_type) -> output_type:
        pass

    def body(ns):
        ns["run"] = run_method

    return types.new_class(
        name="Algorithm[%s, %s]" % (input_type, output_type),
        bases=(Interface,),
        exec_body=body,
    )


class Interface:
    @classmethod
    def is_compatible(cls, other_cls):
        own_methods = _get_annotations(cls, ignore=["generate_cfg", "is_compatible"])

        if not inspect.isclass(other_cls):
            return False

        if issubclass(other_cls, Interface):
            return False

        type_methods = _get_annotations(other_cls)
        return _compatible_annotations(own_methods, type_methods)

    @classmethod
    def generate_cfg(cls, grammar, head):
        symbol = head or Symbol(cls.__name__)
        compatible = []

        for _, other_cls in grammar.namespace.items():
            if cls.is_compatible(other_cls):
                compatible.append(other_cls)

        if not compatible:
            raise ValueError(
                "Cannot find compatible implementations for interface %r" % cls
            )

        return Union(symbol.name, *compatible).generate_cfg(grammar, symbol)


class Distinct:
    def __init__(self, interface: Interface, exceptions=None):
        self.interface = interface
        self.exceptions = exceptions

    def generate_cfg(self, grammar, head):
        symbol = head or Symbol(self.__class__.__name__)
        compatible = []

        for _, other_cls in grammar.namespace.items():
            if other_cls in self.exceptions:
                continue

            if hasattr(other_cls, "__name__") and other_cls.__name__ in self.exceptions:
                continue

            if self.interface.is_compatible(other_cls):
                compatible.append(other_cls)

        if not compatible:
            raise ValueError(
                "Cannot find compatible implementations for interface %r"
                % self.interface
            )

        return Subset(symbol.name, *compatible).generate_cfg(grammar, symbol)


def conforms(type1, type2):
    if inspect.isclass(type1) and inspect.isclass(type2):
        return issubclass(type1, type2)

    if hasattr(type1, "__conforms__") and type1.__conforms__(type2):
        return True

    if hasattr(type2, "__rconforms__") and type2.__rconforms__(type1):
        return True

    return False


def _compatible_annotations(
    methods_if: Mapping[str, inspect.Signature],
    methods_im: Mapping[str, inspect.Signature],
):
    for name, mif in methods_if.items():
        if not name in methods_im:
            return False

        mim = methods_im[name]

        for name, param_if in mif.parameters.items():
            if not name in mim.parameters:
                return False

            param_im = mim.parameters[name]
            ann_if = param_if.annotation

            if ann_if == inspect.Parameter.empty:
                continue

            ann_im = param_im.annotation

            if not conforms(ann_if, ann_im):
                return False

        return_if = mif.return_annotation

        if return_if == inspect.Parameter.empty:
            continue

        return_im = mim.return_annotation

        if not conforms(return_im, return_if):
            return False

    return True


def _get_annotations(clss, ignore=[]):
    """
    Computes the annotations of all public methods in type `clss`.

    ##### Examples

    ```python
    >>> class A:
    ...     def f(self, input: int) -> float:
    ...         pass
    >>> _get_annotations(A)
    {'f': <Signature (self, input:int) -> float>}

    ```
    """
    methods = inspect.getmembers(
        clss, lambda m: inspect.ismethod(m) or inspect.isfunction(m)
    )
    signatures = {
        name: inspect.signature(method)
        for name, method in methods
        if not name.startswith("_")
    }

    for name in ignore:
        signatures.pop(name, None)

    return signatures


def make_list_wrapper(algorithm):
    from autogoal.kb._algorithm import _get_annotations

    input_type, output_type = _get_annotations(algorithm)
    name = f"List[{algorithm.__name__}]"

    def wrap_list(types):
        if isinstance(types, Tuple):
            return Tuple(*(List(t) for t in types.inner))

        return List(types)

    def init_method(self, inner: algorithm):
        self.inner = inner

    def run_method(self, input: wrap_list(input_type)) -> wrap_list(output_type):
        return [self.inner.run(x) for x in xs]

    def repr_method(self):
        return f"{name}(inner={repr(self.inner)})"

    def getattr_method(self, attr):
        return getattr(self.inner, attr)

    def body(ns):
        ns["__init__"] = init_method
        ns["run"] = run_method
        ns["__repr__"] = repr_method
        ns["__getattr__"] = getattr_method

    return types.new_class(name=name, bases=(), exec_body=body)


def build_composite_list(input_type, output_type, depth=1):
    def wrap(t, d):
        if d == 0:
            return t

        return List(wrap(t, d - 1))

    input_wrapper = wrap(input_type, depth)
    output_wrapper = wrap(output_type, depth)

    # name = "ListAlgorithm"  # % (input_wrapper, output_wrapper)
    name = "ListAlgorithm[%s, %s]" % (input_wrapper, output_wrapper)

    def init_method(self, inner: algorithm(input_type, output_type)):
        self.inner = inner

    def run_method(self, input: input_wrapper) -> output_wrapper:
        def wrap_run(xs, d):
            if d == 0:
                return self.inner.run(xs)

            return [wrap_run(x, d - 1) for x in xs]

        return wrap_run(input, depth)

    def repr_method(self):
        return f"{name}(inner={repr(self.inner)})"

    def getattr_method(self, attr):
        return getattr(self.inner, attr)

    def body(ns):
        ns["__init__"] = init_method
        ns["run"] = run_method
        ns["__repr__"] = repr_method
        ns["__getattr__"] = getattr_method

    return types.new_class(name=name, bases=(), exec_body=body)


def build_composite_tuple(index, input_type: "Tuple", output_type: "Tuple"):
    """
    Dynamically generate a class `CompositeAlgorithm` that wraps
    another algorithm to receive a Tuple but pass only one of the
    parameters to the internal algorithm.
    """

    internal_input = input_type.inner[index]
    internal_output = output_type.inner[index]

    # name = "TupleWrapper"
    name = "TupleWrapper[%s, %s]" % (input_type, output_type)

    def init_method(self, inner: algorithm(internal_input, internal_output)):
        self.inner = inner

    def run_method(self, input: input_type) -> output_type:
        elements = list(input)
        elements[index] = self.inner.run(elements[index])
        return tuple(elements)

    def repr_method(self):
        return f"{name}(inner={repr(self.inner)})"

    def getattr_method(self, attr):
        return getattr(self.inner, attr)

    def body(ns):
        ns["__init__"] = init_method
        ns["run"] = run_method
        ns["__repr__"] = repr_method
        ns["__getattr__"] = getattr_method

    return types.new_class(name=name, bases=(), exec_body=body)


class DataType:
    def __init__(self, **tags):
        self.tags = tags

    def get_tag(self, tag):
        return self.tags.get(tag, None)

    def __repr__(self):
        tags = ", ".join(
            f"{key}={value}"
            for key, value in sorted(self.tags.items(), key=lambda t: t[0])
        )
        return f"{self.__class__.__name__}({tags})"

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))

    def __conforms__(self, other):
        return issubclass(self.__class__, other.__class__)


def infer_type(obj):
    """
    Attempts to automatically infer the most precise semantic type for `obj`.

    ##### Parameters

    * `obj`: Object to detect its semantic type.

    ##### Raises

    * `TypeError`: if no valid semantic type was found that matched `obj`.

    ##### Examples

    * Natural language

    ```python
    >>> infer_type("hello")
    Word()
    >>> infer_type("hello world")
    Sentence()
    >>> infer_type("Hello Word. It is raining.")
    Document()

    ```

    * Vectors

    ```
    >>> import numpy as np
    >>> infer_type(np.asarray(["A", "B", "C", "D"]))
    CategoricalVector()
    >>> infer_type(np.asarray([0.0, 1.1, 2.1, 0.2]))
    ContinuousVector()
    >>> infer_type(np.asarray([0, 1, 1, 0]))
    DiscreteVector()

    ```

    * Matrices

    ```
    >>> import numpy as np
    >>> infer_type(np.random.randn(10,10))
    MatrixContinuousDense()

    >>> import scipy.sparse as sp
    >>> infer_type(sp.coo_matrix((10,10)))
    MatrixContinuousSparse()

    ```
    """
    if isinstance(obj, str):
        if " " not in obj:
            return Word()

        if "." not in obj:
            return Sentence()

        return Document()

    if isinstance(obj, list):
        internal_types = set([infer_type(x) for x in obj])

        for test_type in [Document(), Sentence(), Word()]:
            if test_type in internal_types:
                return List(test_type)

    if hasattr(obj, "shape"):
        if len(obj.shape) == 1:
            if isinstance(obj, ndarray):
                if obj.dtype.kind == "U":
                    return CategoricalVector()
                if obj.dtype.kind == "i":
                    return DiscreteVector()
                if obj.dtype.kind == "f":
                    return ContinuousVector()
        if len(obj.shape) == 2:
            if isinstance(obj, spmatrix):
                return MatrixContinuousSparse()
            if isinstance(obj, ndarray):
                return MatrixContinuousDense()

    raise TypeError("Cannot infer type for %r" % obj)


class Text(DataType):
    pass


class Word(Text):
    pass


class Stem(DataType):
    pass


class Sentence(Text):
    pass


class Document(Text):
    pass


class Category(DataType):
    pass


class Vector(DataType):
    pass


class Matrix(DataType):
    pass


class DenseMatrix(Matrix):
    pass


class SparseMatrix(Matrix):
    pass


class ContinuousVector(Vector):
    pass


class DiscreteVector(Vector):
    pass


class CategoricalVector(Vector):
    pass


class MatrixContinuous(Matrix):
    pass


class MatrixContinuousDense(MatrixContinuous, DenseMatrix):
    pass


class MatrixContinuousSparse(MatrixContinuous, SparseMatrix):
    pass


class Entity(DataType):
    pass


class Summary(Document):
    pass


class Sentiment(DataType):
    pass


class Synset(DataType):
    pass


class Postag(DataType):
    pass


class Chunktag(DataType):
    pass


class Tensor3(DataType):
    pass


class Tensor4(DataType):
    pass


class Flags(DataType):
    pass


class List(DataType):
    def __init__(self, inner):
        self.inner = inner
        # super().__init__(**inner.tags)

    def depth(self):
        if not isinstance(self.inner, List):
            return 1

        return 1 + self.inner.depth()

    def __conforms__(self, other):
        return isinstance(other, List) and conforms(self.inner, other.inner)

    def __repr__(self):
        return "List(%r)" % self.inner


class Tuple(DataType):
    def __init__(self, *inner):
        self.inner = inner
        # super().__init__(**inner[0].tags)

    def __repr__(self):
        items = ", ".join(repr(s) for s in self.inner)
        return "Tuple(%s)" % items

    def __conforms__(self, other):
        if not isinstance(other, Tuple):
            return False

        if len(self.inner) != len(other.inner):
            return False

        for x, y in zip(self.inner, other.inner):
            if not conforms(x, y):
                return False

        return True


DATA_TYPES = frozenset(
    [
        Text,
        Word,
        Stem,
        Sentence,
        Document,
        Category,
        Vector,
        Matrix,
        DenseMatrix,
        SparseMatrix,
        ContinuousVector,
        DiscreteVector,
        CategoricalVector,
        MatrixContinuous,
        MatrixContinuousDense,
        MatrixContinuousSparse,
        Entity,
        Summary,
        Sentiment,
        Synset,
        Postag,
        Chunktag,
        Tensor3,
        List,
        Tuple,
        Flags,
    ]
)


def draw_data_hierarchy(output_file):
    """
    Creates an SVG representation of the `DataType` hierarchy,
    for documentation purposes.
    """
    import pydot

    classes = list(DATA_TYPES) + [DataType]

    graph = pydot.Dot(direction="TB")

    for clss in classes:
        graph.add_node(pydot.Node(clss.__name__))

    for clss in classes:
        for base in clss.__bases__:
            if base not in classes:
                continue

            graph.add_edge(pydot.Edge(base.__name__, clss.__name__))

    graph.write(output_file + ".svg", format="svg")
    graph.write(output_file + ".png", format="png")



def find_classes(include=None, exclude=None, modules=None, input=None, output=None):
    import inspect
    import re

    result = []

    if include:
        include = f".*({include}).*"
    else:
        include = r".*"

    if exclude:
        exclude = f".*({exclude}).*"

    if input:
        input = f".*({input}).*"

    if output:
        output = f".*({output}).*"
        
    if modules is None:
        modules = []

        try:
            from autogoal.contrib import sklearn
            modules.append(sklearn)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import nltk
            modules.append(nltk)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import gensim
            modules.append(gensim)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import keras
            modules.append(keras)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import torch
            modules.append(torch)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import spacy
            modules.append(spacy)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import wikipedia
            modules.append(wikipedia)
        except ImportError as e:
            pass

        from autogoal.contrib import wrappers
        modules.append(wrappers)

        from autogoal.contrib import regex
        modules.append(regex)

    for module in modules:
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if not hasattr(cls, 'run'):
                continue

            if cls.__name__.startswith("_"):
                continue

            if not re.match(include, repr(cls)):
                continue

            if exclude is not None and re.match(exclude, repr(cls)):
                continue

            sig = inspect.signature(cls.run)

            if input and not re.match(input, str(sig.parameters["input"].annotation)):
                continue

            if output and not re.match(output, str(sig.return_annotation)):
                continue
            
            result.append(cls)

    return result


__all__ = ['find_classes']



try:
    import wikipedia

    # assert wikipedia.__version__ == "1.4.0"
except:
    print("(!) Code in `autogoal.contrib.wikipedia` requires `wikipedia==1.4.0`.")
    print("(!) You can install it with `pip install autogoal[wikipedia]`.")
    raise


from ._base import (
    WikipediaContainsWord,
    WikipediaContainsWordSpanish,
    WikipediaSummary,
    WikipediaSummarySpanish,
)



import warnings
import inspect
import re
import numpy as np
import scipy.sparse as sp

from autogoal import kb
from autogoal.contrib.sklearn._utils import is_matrix_continuous_dense,\
                                            is_matrix_continuous_sparse,\
                                            is_categorical,\
                                            is_continuous,\
                                            is_string_list

DATA_TYPE_EXAMPLES = {
    kb.Postag():("lorem", "ipsum"), # (str, str) Tagged token
    kb.List(kb.Postag()):[("lorem", "ipsum")] * 10, # [(str, str), (str, str)] List of tagged tokens
    kb.List(kb.List(kb.Postag())):[[("lorem", "ipsum")] * 2], # [[(str, str), (str, str)], [(str, str), (str, str)]] List of Tagged Sentences
    kb.Chunktag():(("lorem", "ipsum"),"ipsum"), # ((str, str), str) IOB Tagged token
    kb.List(kb.Chunktag()):[(("lorem", "ipsum"),"ipsum")] * 10, # [((str, str), str), ((str, str), str)] List of IOB Tagged token
    kb.List(kb.List(kb.Chunktag())):[[(("lorem", "ipsum"),"ipsum")] * 2], # [[((str, str), str), ((str, str), str)], [((str, str), str), ((str, str), str)]] List of IOB Tagged Sentences
    kb.Stem():"ips",
    kb.Word():"ipsum",
    kb.Sentence():"It is the best of all movies.",
    kb.Document():"It is the best of all movies. I actually love that action scene.",
    kb.MatrixContinuousDense(): np.random.rand(10, 10),
    kb.MatrixContinuousSparse(): sp.rand(10, 10),
    kb.CategoricalVector(): np.asarray(["A"] * 5 + ["B"] * 5),
    kb.ContinuousVector(): np.random.rand(10),
    kb.DiscreteVector(): np.random.randint(0, 10, (10,), dtype=int),
    kb.List(kb.Word()):["ipsu", "lorem"],
    kb.List(kb.Document()): ["abc ipsu lorem say hello", "ipsum lorem", "abc"] * 2,
    kb.List(kb.List(kb.Stem())): [["abc", "ipsu", "lorem"] * 10],
    kb.List(kb.List(kb.Word())): [["abc", "ipsu", "lorem"] * 10],
    kb.List(kb.List(kb.Sentence())): [["abc a sentence lorem"], ["ipsum lorem"], ["abc"]]
}

def is_algorithm(cls, verbose=False):
    if _is_classifier(cls):
        return "classifier"

    if _is_clusterer(cls):
        return "clusterer"

    if _is_sent_tokenizer(cls):
        return "sent_tokenizer"

    if _is_word_tokenizer(cls):
        return "word_tokenizer"

    if _is_lemmatizer(cls):
        return "lemmatizer"

    if _is_stemmer(cls):
        return "stemmer"

    if _is_word_embbeder(cls):
        return "word_embbeder"

    if _is_doc_embbeder(cls):
        return "doc_embbeder"
    
    if is_pretrained_tagger(cls)[0]:
        return "trained_tagger"
    
    if _is_tagger(cls):
        return "tagger"
    
    return False

def _is_algorithm(cls, verbose = False):
    return  _is_stemmer(cls, verbose) or\
            _is_lemmatizer(cls, verbose) or\
            _is_word_tokenizer(cls, verbose) or\
            _is_sent_tokenizer(cls, verbose) or\
            _is_clusterer(cls, verbose) or\
            _is_classifier(cls, verbose) or\
            _is_word_embbeder(cls, verbose) or\
            _is_doc_embbeder(cls, verbose) or\
            _is_tagger(cls, verbose)

def _is_stemmer(cls, verbose=False):
    if hasattr(cls, "stem"):
        return True
    return False

def _is_lemmatizer(cls, verbose=False):
    if hasattr(cls, "lemmatize"):
        return True
    return False

def _is_word_tokenizer(cls, verbose=False):
    if (hasattr(cls, "tokenize") or hasattr(cls, "word_tokenize")):
        return True
    return False

def _is_sent_tokenizer(cls, verbose=False):
    if hasattr(cls, "sent_tokenize") or (hasattr(cls, "tokenize")):
        return True
    return False

def _is_clusterer(cls, verbose=False):
    if (hasattr(cls, "classify") and hasattr(cls, "cluster")):
        return True
    return False

def _is_classifier(cls, verbose = False):
    if (hasattr(cls, "classify") and hasattr(cls, "train")):
        return True
    return False

def _is_word_embbeder(cls, verbose = False):
    if (hasattr(cls, "build_vocab") and hasattr(cls, "train") and hasattr(cls, "wv")):
        return True
    return False

def _is_doc_embbeder(cls, verbose = False):
    if (hasattr(cls, "build_vocab") and hasattr(cls, "train") and hasattr(cls, "infer_vector")):
        return True
    return False

def _is_tagger(cls, verbose = False):
    if hasattr(cls, "tag"):
        return True
    return False

def _is_trained_tagger(cls, verbose = False):
    if hasattr(cls, "train") and _is_tagger(cls, verbose):
        return True
    return False

def _is_chunker(cls, verbose = False):
    if hasattr(cls, "parse"):
        return True
    return False
    

def is_stemmer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk stemmer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression
    >>> from nltk.stem import Cistem
    >>> is_stemmer(Cistem)
    (True, (Word(), Stem()))
    >>> is_stemmer(LogisticRegression)
    (False, None)

    """
    if not _is_stemmer(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.Stem()

    for input_type in [kb.Word()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            stemmer = cls()
            y = stemmer.stem(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None

def is_lemmatizer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk lemmatizer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression
    >>> from nltk.stem import WordNetLemmatizer
    >>> is_lemmatizer(WordNetLemmatizer)
    (True, (Word(), Stem()))
    >>> is_lemmatizer(LogisticRegression)
    (False, None)

    """
    if not _is_lemmatizer(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.Stem()
    for input_type in [kb.Word()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            lemmatizer = cls()
            y = lemmatizer.lemmatize(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None

def is_word_tokenizer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk word tokenizer algorithm.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression
    >>> from nltk.tokenize import TweetTokenizer
    >>> is_word_tokenizer(TweetTokenizer)
    (True, (Sentence(), List(Word())))
    >>> is_word_tokenizer(LogisticRegression)
    (False, None)

    """
    if not _is_word_tokenizer(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.List(kb.Word())

    for input_type in [kb.Sentence()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            tokenizer = cls()
            y = tokenizer.tokenize(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None

def is_sent_tokenizer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk sentence tokenizer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression
    >>> from nltk.tokenize import PunktSentenceTokenizer
    >>> is_sent_tokenizer(PunktSentenceTokenizer)
    (True, (Document(), List(Sentence())))
    >>> is_sent_tokenizer(LogisticRegression)
    (False, None)

    """
    if not _is_sent_tokenizer(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.List(kb.Sentence())

    for input_type in [kb.Document()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            tokenizer = cls()
            y = tokenizer.tokenize(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None

def is_clusterer(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk clusterer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from sklearn.linear_model import LogisticRegression
    >>> from nltk.cluster import GAAClusterer
    >>> is_clusterer(GAAClusterer)
    (True, (MatrixContinuousDense(), CategoricalVector()))
    >>> is_clusterer(LogisticRegression)
    (False, None)

    """
    if not _is_clusterer(cls, verbose=verbose):
        return False, None

    inputs = []

    for input_type in [kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]
            y = DATA_TYPE_EXAMPLES[kb.CategoricalVector()]

            clusterer = cls()
            clusterer.cluster(X)
            y = [clusterer.classify(x) for x in X]

            assert is_categorical(y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, kb.CategoricalVector())
    else:
        return False, None

def is_classifier(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk classifier.
    If True, returns the valid (input, output) types.
    """
    if not _is_algorithm(cls, verbose=verbose):
        return False, None

    inputs = []

    #TODO: Fix somehow compatibility with nltk classifiers
    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, kb.ContinuousVector())
    else:
        return False, None

def is_tagger(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk pos tagger.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from nltk.tag import AffixTagger
    >>> from nltk.tokenize import PunktSentenceTokenizer
    >>> is_tagger(AffixTagger)
    (True, (List(List(Word())), List(List(Postag()))))
    >>> is_tagger(LogisticRegression)
    (False, None)

    """
    if not _is_tagger(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.List(kb.List(kb.Postag()))

    for input_type in [kb.List(kb.List(kb.Word()))]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            X_train = [[(word,word) for word in sentence] for sentence in X]
            
            tagger = cls(train=X_train)
            # tagger = cls()
            y = tagger.tag_sents(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        is_ptt = is_pretrained_tagger(cls, verbose)
        is_ckr = is_chunker(cls, verbose)
        
        if is_ptt[0]:
            return is_ptt
        if is_ckr[0]:
            return is_ckr
        return False, None

def is_chunker(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk chunker.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from nltk.chunk.named_entity import NEChunkParserTagger
    >>> from nltk.tokenize import PunktSentenceTokenizer
    >>> is_chunker(NEChunkParserTagger)
    (True, (List(List(Tuple(Word(), Word()))), List(List(Tuple(Tuple(Word(), Word()), Word())))))
    >>> is_chunker(PunktSentenceTokenizer)
    (False, None)

    """
    if not _is_tagger(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.List(kb.List(kb.Chunktag()))

    for input_type in [kb.List(kb.List(kb.Postag()))]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]

            X_train = [[((word, postag), postag) for word, postag in sentence] for sentence in X]
            
            chunker = cls(train=X_train)
            y = chunker.tag_sents(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None

def is_pretrained_tagger(cls, verbose=False):
    """Determine if `cls` corresponds to something that resembles an nltk sentence tokenizer.
    If True, returns the valid (input, output) types.

    Examples:

    >>> from nltk.tag import AffixTagger
    >>> from nltk.tag.perceptron import PerceptronTagger
    >>> is_pretrained_tagger(PerceptronTagger)
    (True, (List(Word()), List(Tuple(Word(), Word()))))
    >>> is_pretrained_tagger(AffixTagger)
    (False, None)

    """
    if not _is_tagger(cls, verbose=verbose):
        return False, None

    inputs = []
    output = kb.List(kb.Postag())

    for input_type in [kb.List(kb.Word())]:
        try:
            X = DATA_TYPE_EXAMPLES[input_type]
            
            tagger = cls()
            
            y = tagger.tag(X)

            assert DATA_RESOLVERS[output](y)
            inputs.append(input_type)
        except Exception as e:
            if verbose:
                warnings.warn(str(e))

    inputs = combine_types(*inputs)

    if inputs:
        return True, (inputs, output)
    else:
        return False, None
    
def is_data_type(X, data_type):
    return DATA_RESOLVERS[data_type](X)

IO_TYPE_HANDLER = [
    is_stemmer,
    is_lemmatizer,
    is_word_tokenizer,
    is_sent_tokenizer,
    is_clusterer,
    is_classifier,
    is_tagger,
    # is_word_embbeder,
    # is_doc_embbeder
]

def get_input_output(cls, verbose=False):
    for func in IO_TYPE_HANDLER:
        matches, types = func(cls, verbose=verbose)
        if matches:
            return types

    return None, None

def combine_types(*types):
    if len(types) == 1:
        return types[0]

    types = set(types)

    if types == {kb.MatrixContinuousDense(), kb.MatrixContinuousSparse()}:
        return kb.MatrixContinuous()

    return None

def is_word_list(obj):
    """Determines if `obj` is a sequence of sequence of strings.

    Examples:

    >>> is_word_list(['hello', 'world'])
    True
    >>> is_word_list(np.random.rand(10))
    False

    """
    try:
        oset = set()
        for word in obj:
            if len(word.split()) > 1:
                return False
            oset.add(word)
        return len(oset) > 0.1 * len(obj) and all(isinstance(x, str) for x in oset)
    except:
        return False

def is_word_list_list(obj):
    """Determines if `obj` is a sequence of sequence of strings.

    Examples:

    >>> is_word_list_list([['hello'], ['another'], ['word']])
    True
    >>> is_word_list_list(np.random.rand(10))
    False

    """
    try:
        oset = set()
        for sent in obj:
            for word in sent:
                if len(word.split()) > 1:
                    return False
                oset.add(word)
        return len(oset) > 0.1 * len(obj) and all(isinstance(x, str) for x in oset)
    except:
        return False

def is_word(obj):
    """Determines if `obj` is a sequence of sequence of strings.

    Examples:

    >>> is_word('hello')
    True
    >>> is_word(np.random.rand(10))
    False

    """
    try:
        return isinstance(obj, str) and len(obj.split()) == 1
    except:
        return False

def is_sentence(obj):
    """Determines if `obj` is a sentence strings.

    Examples:

    >>> is_sentence('hello world')
    True
    >>> is_word(np.random.rand(10))
    False

    """
    try:
        return isinstance(obj, str) and len(obj.split()) > 1
    except:
        return False

def is_sentence_list(obj):
    try:
        return all([is_sentence(x) for x in obj])
    except:
        return False

def is_text_list_list(obj):
    """Determines if `obj` is a sequence of sequence of strings.

    Examples:

    >>> is_text_list_list([['hello', 'world'], ['another'], ['sentence']])
    True
    >>> is_text_list_list(np.random.rand(10))
    False

    """
    try:
        oset = set()

        for sent in obj:
            for text in sent:
                oset.add(text)

        return len(oset) > 0.1 * len(obj) and all(isinstance(x, str) for x in oset)
    except:
        return False

def is_tag(obj):
    """Determines if `obj` is a tuple of two strings.

    Examples:

    >>> is_tag(('hello', 'yes'))
    True
    >>> is_tag(('hi', 22))
    False

    """
    try:
        return isinstance(obj, tuple) and len(obj) == 2 and all((isinstance(x, str) or x == None) for x in obj)
    except:
        return False
    
def is_tag_list(obj):
    """Determines if `obj` is a list of tuple of two strings.

    Examples:

    >>> is_tag_list([('hello', 'yes'), ('how', 'is')])
    True
    >>> is_tag_list(('hello', 'yes'))
    False

    """
    try:
        return isinstance(obj, list) and all(is_tag(x) for x in obj)
    except:
        return False

def is_tagged_sentence_list(obj):
    """Determines if `obj` is a list(list(tuple(str, str)))

    Examples:

    >>> is_tagged_sentence_list([[('hello', 'yes'), ('how', 'is')]])
    True
    >>> is_tagged_sentence_list([('hello', 'yes')])
    False

    """
    try:
        return isinstance(obj, list) and all(is_tag_list(x) for x in obj)
    except:
        return False


def is_chunk(obj):
    """Determines if `obj` is a tuple(tuple(str, str), str).

    Examples:

    >>> is_chunk((('hello', 'yes'),'how'))
    True
    >>> is_chunk(('hello', 'yes'))
    False

    """
    try:
        return isinstance(obj, tuple) and len(obj) == 2 and is_tag(obj[0]) and isinstance(obj[1], str)
    except:
        return False

def is_chunk_list(obj):
    """Determines if `obj` is a list(tuple(tuple(str, str), str)).

    Examples:

    >>> is_chunk_list([(('hello', 'yes'),'how'), (('are', 'you'), 'today')])
    True
    >>> is_chunk_list([('hello', 'yes'), ('how', 'are')])
    False

    """
    try:
        return isinstance(obj, list) and all(is_chunk(x) for x in obj)
    except:
        return False

def is_chunked_sentence_list(obj):
    """Determines if `obj` is a list(list(tuple(tuple(str, str), str))).

    Examples:

    >>> is_chunked_sentence_list([[(('hello', 'yes'),'how'), (('are', 'you'), 'today')]])
    True
    >>> is_chunked_sentence_list([('hello', 'yes'), ('how', 'are')])
    False

    """
    try:
        return isinstance(obj, list) and all(is_chunk_list(x) for x in obj)
    except:
        return False


DATA_RESOLVERS = {
    kb.Postag():is_tag,
    kb.List(kb.Postag()):is_tag_list,
    kb.List(kb.List(kb.Postag())):is_tagged_sentence_list,
    kb.Chunktag:is_chunk,
    kb.List(kb.Chunktag()):is_chunk_list,
    kb.List(kb.List(kb.Chunktag())):is_chunked_sentence_list,
    kb.Stem():is_word,
    kb.Word():is_word,
    kb.Sentence():is_sentence,
    kb.Document():is_sentence,
    kb.MatrixContinuousDense(): is_matrix_continuous_dense,
    kb.MatrixContinuousSparse(): is_matrix_continuous_sparse,
    kb.CategoricalVector(): is_categorical,
    kb.ContinuousVector(): is_continuous,
    kb.List(kb.Word()): is_word_list,
    kb.List(kb.Sentence()): is_sentence_list,
    kb.List(kb.List(kb.Stem())): is_word_list_list,
    kb.List(kb.List(kb.Word())): is_word_list_list,
}



from autogoal.search import RandomSearch, PESearch
from autogoal.kb import (
    build_pipeline_graph,
    build_pipelines,
    CategoricalVector,
    List,
    Word,
    MatrixContinuous,
    Tuple,
    infer_type,
    Postag,
    Chunktag
)

from autogoal.ml.metrics import accuracy
from autogoal.sampling import ReplaySampler
from autogoal.contrib import find_classes
# from autogoal.ml._metalearning import DatasetFeatureLogger

import numpy as np
import random
import statistics
import pickle



class AutoML:
    """
    Predefined pipeline search with automatic type inference.

    An `AutoML` instance represents a general-purpose machine learning
    algorithm, that can be applied to any input and output.
    """
    def __init__(
        self,
        input=None,
        output=None,
        random_state=None,
        search_algorithm=PESearch,
        search_kwargs={},
        search_iterations=100,
        include_filter=".*",
        exclude_filter=None,
        validation_split=0.3,
        errors="warn",
        cross_validation="median",
        cross_validation_steps=3,
        registry=None,
        score_metric=None,
        metalearning_log=False,
    ):
        self.input = input
        self.output = output
        self.search_algorithm = search_algorithm
        self.search_kwargs = search_kwargs
        self.search_iterations = search_iterations
        self.include_filter = include_filter
        self.exclude_filter = exclude_filter
        self.validation_split = validation_split
        self.errors = errors
        self.cross_validation = cross_validation
        self.cross_validation_steps = cross_validation_steps
        self.registry = registry
        self.random_state = random_state
        self.score_metric = score_metric or accuracy
        self.metalearning_log = metalearning_log

        if random_state:
            np.random.seed(random_state)

    def _make_pipeline_builder(self):
        registry = self.registry or find_classes(
            include=self.include_filter, exclude=self.exclude_filter
        )

        return build_pipelines(
            input=Tuple(self.input, self.output),
            output=self.output,
            registry=registry,
        )

    def fit(self, X, y, **kwargs):
        self.input = self._input_type(X)
        self.output = self._output_type(y)

        if self.metalearning_log:
            raise NotImplementedError("Metalearning is not ready yet")

            loggers = kwargs.get('logger', [])
            loggers.append(DatasetFeatureLogger(X, y, problem_features=dict(
                input=repr(self.input),
                output=repr(self.output),
                metric=self.score_metric.__name__,
            ), environment_features=dict(
                memory_limit=self.search_kwargs.get('memory_limit'),
                search_timeout=self.search_kwargs.get('search_timeout'),
                evaluation_timeout=self.search_kwargs.get('evaluation_timeout'),
            )))
            kwargs['logger'] = loggers

        search = self.search_algorithm(
            self._make_pipeline_builder(),
            self._make_fitness_fn(X, y),
            random_state=self.random_state,
            errors=self.errors,
            **self.search_kwargs,
        )

        self.best_pipeline_, self.best_score_ = search.run(
            self.search_iterations, **kwargs
        )

        self.fit_pipeline(X, y)

    def fit_pipeline(self, X, y):
        if not hasattr(self, 'best_pipeline_'):
            raise TypeError("You have to call `fit()` first.")

        self.best_pipeline_.send("train")
        self.best_pipeline_.run((X, y))
        self.best_pipeline_.send("eval")

    def save_pipeline(self, fp):
        """
        Saves the state of the best pipeline.
        You are responsible for opening and closing the stream.
        """
        if not hasattr(self, 'best_pipeline_'):
            raise TypeError("You have to call `fit()` first.")

        self.best_pipeline_.sampler_.replay().save(fp)
        pickle.Pickler(fp).dump((self.input, self.output))

    def load_pipeline(self, fp):
        """
        Loads the state of the best pipeline and retrains.
        You are responsible for opening and closing the stream.

        After calling load, the best pipeline is **not** trained.
        You need to retrain it by calling `fit_pipeline(X, y)`.
        """
        sampler = ReplaySampler.load(fp)
        self.input, self.output = pickle.Unpickler(fp).load()
        self.best_pipeline_ = self._make_pipeline_builder()(sampler)

    def score(self, X, y):
        y_pred = self.best_pipeline_.run((X, np.zeros_like(y)))
        return self.score_metric(y, y_pred)

    def _input_type(self, X):
        return self.input or infer_type(X)

    def _output_type(self, y):
        return self.output or infer_type(y)

    def _make_fitness_fn(self, X, y):
        y = np.asarray(y)

        def fitness_fn(pipeline):
            scores = []

            for _ in range(self.cross_validation_steps):
                len_x = len(X) if isinstance(X, list) else X.shape[0]
                indices = np.arange(0, len_x)
                np.random.shuffle(indices)
                split_index = int(self.validation_split * len(indices))
                train_indices = indices[:-split_index]
                test_indices = indices[-split_index:]

                if isinstance(X, list):
                    X_train, y_train, X_test, y_test = (
                        [X[i] for i in train_indices],
                        y[train_indices],
                        [X[i] for i in test_indices],
                        y[test_indices],
                    )
                else:
                    X_train, y_train, X_test, y_test = (
                        X[train_indices],
                        y[train_indices],
                        X[test_indices],
                        y[test_indices],
                    )

                pipeline.send("train")
                pipeline.run((X_train, y_train))
                pipeline.send("eval")
                y_pred = pipeline.run((X_test, np.zeros_like(y_test)))
                scores.append(self.score_metric(y_test, y_pred))

            return getattr(statistics, self.cross_validation)(scores)

        return fitness_fn

    def predict(self, X):
        return self.best_pipeline_.run((X, [None] * len(X)))



try:
    from tensorflow import keras

    # assert keras.__version__ == "2.3.1"
except:
    print("(!) Code in `autogoal.contrib.keras` requires `keras==2.3.1`.")
    print("(!) You can install it with `pip install autogoal[keras]`.")
    raise


from ._base import (
    KerasClassifier,
    KerasSequenceClassifier,
    KerasSequenceTagger,
    KerasImageClassifier,
    KerasImagePreprocessor,
)
from ._grammars import build_grammar



from sklearn.feature_extraction.text import CountVectorizer as _CountVectorizer
from sklearn.feature_extraction import DictVectorizer
from sklearn_crfsuite import CRF

from autogoal.contrib.sklearn._builder import SklearnTransformer, SklearnEstimator
from autogoal.kb import (
    List,
    Sentence,
    MatrixContinuousSparse,
    MatrixContinuousDense,
    algorithm,
    Word,
    Stem,
    Flags,
    Vector,
    Category,
    Tuple,
)
from autogoal.grammar import Boolean, Categorical, Discrete, Continuous
from autogoal.utils import nice_repr


@nice_repr
class CountVectorizerNoTokenize(_CountVectorizer, SklearnTransformer):
    def __init__(
        self,
        lowercase: Boolean(),
        stopwords_remove: Boolean(),
        binary: Boolean(),
        inner_tokenizer: algorithm(Sentence(), List(Word())),
        inner_stemmer: algorithm(Word(), Stem()),
        inner_stopwords: algorithm(List(Word()), List(Word())),
    ):
        self.stopwords_remove = stopwords_remove
        self.inner_tokenizer = inner_tokenizer
        self.inner_stemmer = inner_stemmer
        self.inner_stopwords = inner_stopwords

        SklearnTransformer.__init__(self)
        _CountVectorizer.__init__(self, lowercase=lowercase, binary=binary)

    def build_tokenizer(self):
        def func(sentence):
            tokens = self.inner_tokenizer.run(sentence)
            tokens = (
                self.inner_stopwords.run(sentence) if self.stopwords_remove else tokens
            )
            return [self.inner_stemmer.run(token) for token in tokens]

        return func

    def run(self, input: List(Sentence())) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


class _FlagsVectorizer(SklearnTransformer):
    def __init__(self, sparse):
        self.vectorizer = DictVectorizer(sparse=sparse)
        super().__init__()

    def fit_transform(self, X, y=None):
        return self.vectorizer.fit_transform(X)

    def transform(self, X, y=None):
        return self.vectorizer.transform(X, y=y)

@nice_repr
class FlagsSparseVectorizer(_FlagsVectorizer):
    def __init__(self):
        super().__init__(sparse=True)

    def run(self, input: List(Flags())) -> MatrixContinuousSparse():
        return super().run(input)


@nice_repr
class FlagsDenseVectorizer(_FlagsVectorizer):
    def __init__(self):
        super().__init__(sparse=False)

    def run(self, input: List(Flags())) -> MatrixContinuousDense():
        return super().run(input)


@nice_repr
class CRFTagger(CRF, SklearnEstimator):
    def __init__(self, algorithm: Categorical('lbfgs', 'l2sgd', 'ap', 'pa', 'arow')) -> None:
        SklearnEstimator.__init__(self)
        super().__init__(algorithm=algorithm)

    def run(self, input: Tuple(List(List(Flags())), List(List(Category())))) -> List(List(Category())):
        return SklearnEstimator.run(self, input)


__all__ = [
    "CountVectorizerNoTokenize",
    "FlagsSparseVectorizer",
    "FlagsDenseVectorizer",
    "CRFTagger"
]


import random

from ._base import SearchAlgorithm
from autogoal.sampling import Sampler


class RandomSearch(SearchAlgorithm):
    def __init__(self, *args, random_state: int = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._sampler = Sampler(random_state=random_state)

    def _build_sampler(self):
        return self._sampler



import mmap
import functools


import pickle, json, csv, os, shutil


class PersistentDict(dict):
    """
    Persistent dictionary with an API compatible with shelve and anydbm.

    Taken from <https://code.activestate.com/recipes/576642/>

    The dict is kept in memory, so the dictionary operations run as fast as
    a regular dictionary.

    Write to disk is delayed until close or sync (similar to gdbm's fast mode).

    Input file format is automatically discovered.
    Output file format is selectable between pickle, json, and csv.
    All three serialization formats are backed by fast C implementations.

    """

    def __init__(self, filename, flag="c", mode=None, format="pickle", *args, **kwds):
        self.flag = flag  # r=readonly, c=create, or n=new
        self.mode = mode  # None or an octal triple like 0644
        self.format = format  # 'csv', 'json', or 'pickle'
        self.filename = filename
        if flag != "n" and os.access(filename, os.R_OK):
            fileobj = open(filename, "rb" if format == "pickle" else "r")
            with fileobj:
                self.load(fileobj)
        dict.__init__(self, *args, **kwds)

    def sync(self):
        "Write dict to disk"
        if self.flag == "r":
            return
        filename = self.filename
        tempname = filename + ".tmp"
        fileobj = open(tempname, "wb" if self.format == "pickle" else "w")
        try:
            self.dump(fileobj)
        except Exception:
            os.remove(tempname)
            raise
        finally:
            fileobj.close()
        shutil.move(tempname, self.filename)  # atomic commit
        if self.mode is not None:
            os.chmod(self.filename, self.mode)

    def close(self):
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def dump(self, fileobj):
        if self.format == "csv":
            csv.writer(fileobj).writerows(self.items())
        elif self.format == "json":
            json.dump(self, fileobj, separators=(",", ":"))
        elif self.format == "pickle":
            pickle.dump(dict(self), fileobj, 2)
        else:
            raise NotImplementedError("Unknown format: " + repr(self.format))

    def load(self, fileobj):
        # try formats from most restrictive to least restrictive
        for loader in (pickle.load, json.load, csv.reader):
            fileobj.seek(0)
            try:
                return self.update(loader(fileobj))
            except Exception:
                pass
        raise ValueError("File not in a supported format")


from pathlib import Path


class CacheManager:
    _instance = None

    def __init__(self):
        self.cache = PersistentDict(str(Path(__file__).parent / "cache.pickle"))

    @staticmethod
    def get(name: str, func):
        instance = CacheManager.instance()

        if name not in instance.cache:
            print("Creating cached object '%s'" % name)
            instance.cache[name] = func()
            instance.cache.sync()

        return instance.cache[name]

    @staticmethod
    def instance() -> "CacheManager":
        if CacheManager._instance is None:
            CacheManager._instance = CacheManager()

        return CacheManager._instance


def cached_run(func):
    @functools.wraps(func)
    def run(self, input):
        if not hasattr(input, "__cached_id__"):
            return func(input)

        cached_id = input.__cached_id__

    return run



import inspect
import numpy as np

from autogoal.kb import CategoricalVector
from autogoal.kb import conforms


METRICS = []


def register_metric(func):
    METRICS.append(func)
    return func


def find_metric(*types):
    for metric_func in METRICS:
        signature = inspect.signature(metric_func)

        if len(types) != len(signature.parameters):
            break

        for type_if, type_an in zip(types, signature.parameters):
            if not conforms(type_an.annotation, type_if):
                break

        return metric_func

    raise ValueError("No metric found for types: %r" % types)


def accuracy(ytrue: CategoricalVector, ypred: CategoricalVector) -> float:
    return np.mean([1 if yt == yp else 0 for yt,yp in zip(ytrue, ypred)])



from autogoal.datasets import download, datapath
import pickle
import numpy as np


def load(training_batches=5):
    """
    Load the CIFAR-10 dataset

    ##### Parameters

    * 'training_batches': maximum number of batches to load for training, 
      each batch has 10,000 examples (min=`1`, max=`5`, default=`5`).

    ##### Examples

    >>> X_train, y_train, X_test, y_test = load(training_batches=5)
    >>> X_train.shape
    (50000, 32, 32, 3)
    >>> len(y_train)
    50000
    >>> X_test.shape
    (10000, 32, 32, 3)
    >>> len(y_test)
    10000
    >>> y_train[0]
    6

    """
    download("cifar10")

    X_train = []
    y_train = []

    for i in range(1, training_batches + 1):
        batch = datapath('cifar10') / f'data_batch_{i}'

        with open(batch, 'rb') as fp:
            data = pickle.load(fp, encoding='bytes')
            X_train.append(data[b'data'])
            y_train.extend(data[b'labels'])

    X_train = np.vstack(X_train)
    X_train = np.reshape(X_train, (-1, 3, 32, 32)).transpose(0, 2, 3, 1)

    test_batch = datapath('cifar10') / 'test_batch'

    with open(test_batch, 'rb') as fp:
        data = pickle.load(fp, encoding='bytes')
        X_test, y_test = data[b'data'], data[b'labels']
        X_test = np.reshape(X_test, (-1, 3, 32, 32)).transpose(0, 2, 3, 1)

    return X_train, y_train, X_test, y_test



import numpy as np
import os
from autogoal.datasets import datapath, download
from sklearn.feature_extraction import DictVectorizer


def load(max_examples=None):
    """
    Loads train and valid datasets from [Shuttle uci dataset](https://archive.ics.uci.edu/ml/datasets/Statlog+(Shuttle)).

    ##### Examples

    ```python
    >>> X_train, y_train, X_valid, y_valid = load()
    >>> X_train.shape, X_valid.shape
    ((43500, 9), (14500, 9))
    >>> len(y_train), len(y_valid)
    (43500, 14500)

    ```
    """

    try:
        download("shuttle")
    except:
        print("Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry")
        raise

    path = str(datapath(os.path.dirname(os.path.abspath(__file__)))) + "/data/shuttle"
    train_data = open(os.path.join(path, "shuttle.trn"), "r")
    test_data = open(os.path.join(path, "shuttle.tst"), "r")

    X_train = []
    X_test = []
    y_train = []
    y_test = []

    for i in train_data.readlines():
        clean_line = i.strip().split()

        temp = {}
        temp["1"] = int(clean_line[0])
        temp["2"] = int(clean_line[1])
        temp["3"] = int(clean_line[2])
        temp["4"] = int(clean_line[3])
        temp["5"] = int(clean_line[4])
        temp["6"] = int(clean_line[5])
        temp["7"] = int(clean_line[6])
        temp["8"] = int(clean_line[7])
        temp["9"] = int(clean_line[8])

        X_train.append(temp)
        y_train.append(clean_line[9])

        if max_examples and len(X_train) >= max_examples:
            break

    for i in test_data.readlines():
        clean_line = i.strip().split()

        temp = {}
        temp["1"] = int(clean_line[0])
        temp["2"] = int(clean_line[1])
        temp["3"] = int(clean_line[2])
        temp["4"] = int(clean_line[3])
        temp["5"] = int(clean_line[4])
        temp["6"] = int(clean_line[5])
        temp["7"] = int(clean_line[6])
        temp["8"] = int(clean_line[7])
        temp["9"] = int(clean_line[8])

        X_test.append(temp)
        y_test.append(clean_line[9])

        if max_examples and len(X_test) >= max_examples:
            break

    X_train, y_train = _load_onehot(X_train, y_train)
    X_test, y_test = _load_onehot(X_test, y_test)

    return X_train, y_train, X_test, y_test


def _load_onehot(X, y):
    vec = DictVectorizer(sparse=False)

    return vec.fit_transform(X), np.asarray(y)



import numpy as np
import os
from autogoal.datasets import datapath, download
from sklearn.feature_extraction import DictVectorizer

def load():
    """
    Loads corpora from [Yeast uci dataset](https://archive.ics.uci.edu/ml/datasets/Yeast).

    ##### Examples

    ```python
    >>> X, y = load()
    >>> X.shape
    (1484, 8)
    >>> len(y)
    1484

    ```
    """

    try:
        download("yeast")
    except:
        print("Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry")
        raise

    path = str(datapath(os.path.dirname(os.path.abspath(__file__)))) + "/data/yeast"
    f = open(os.path.join(path, "yeast.data"), "r")

    X = []
    y = []

    for i in f:
        clean_line = i.strip().split()
        temp = {}
        temp["1"] = float(clean_line[1])
        temp["2"] = float(clean_line[2])
        temp["3"] = float(clean_line[3])
        temp["4"] = float(clean_line[4])
        temp["5"] = float(clean_line[5])
        temp["6"] = float(clean_line[6])
        temp["7"] = float(clean_line[7])
        temp["8"] = float(clean_line[8])

        X.append(temp)
        y.append(clean_line[9])

    return _load_onehot(X, y)


def _load_onehot(X, y):
    vec = DictVectorizer(sparse=False)

    return vec.fit_transform(X), np.asarray(y)
